/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

Effect main class

General algorithm is:
1. Delay input into input buffer (_x0)
2. If we can interpolate the params or we're forced to output data FFT data from _x0 to _x1
3. run an effect from FFTFx()
4. IFFT data from _x1 to _x2
5. mix data from _x2 into _x3 (output buffer), doing a cross fade with any overlap of data
   already in there

History
	Date       Version    Programmer         Comments
	16/2/03    -          Darrell Tam		 Created buzz version
    01/10/05   1.0        Darrell Tam        Ported to VST

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/

#include <stdio.h>
#include <string.h>
#include <dt/cplx.h>
#include <dt/WrapProcessFloatVec.h>

#include "rfftw_float.h"
#include "DtBlkFx.hpp"
#include "AEffEditor.hpp"
#include "Gui.h"
#include "FFTFx.h"

#define _DBG
#define DBG_FILE_NAME "c:\\blkfx.txt"
#include "Debug.h"

using namespace std;

//-------------------------------------------------------------------------------------------------

DtBlkFx::DtBlkFx (audioMasterCallback audioMaster)
	: AudioEffectX (audioMaster, /*kNumPrograms*/g_num_blk_fx_presets, /*kNumParams*/BlkFxParam::TOTAL_NUM)
{	
	DBG("DtBlkFx constructor");
	unsigned i;

	_chunk_data.tag = ChunkData::TAG;
	_chunk_data.version = ChunkData::VERS0;		// version 0 chunk

	// get buffer space
	_x0_sz = 8*44100;				// length of buffer is arbitrary
	_x0_force_out_sz = _x0_sz - MAX_FFT_BLK_SZ; // force output when x0 contains this much data
    _x0.resize(_x0_sz+MAX_FFT_BLK_SZ);	// input buffer with extra space at end to unwrap data for processing
    _x1.resize(MAX_FFT_BLK_SZ+1);	// allow 1 extra for imaginary part of DC (which is always 0)
    _x2.resize(MAX_FFT_BLK_SZ+1);	// inverse FFT
	_x3.resize(8*44100);			// output buffer (length is arbitrary)

	_max_delay_n = _x3.size()-MAX_FFT_BLK_SZ-2048;

	// allocate 10k per effect
	for(i = 0; i < BlkFxParam::NUM_FX_SETS; i++)
		_fx_params[i].data.resize(10000);

	// put in arbitrary initial params - they should be overwritten by first param update
	_params.put(0, BlkFxParam::DELAY, 16.0f/255.0f);
	_params.put(0, BlkFxParam::FFT_LEN, BlkFxParam::getFFTLenParam(5));
	_params.put(0, BlkFxParam::OVERLAP, 0.35f);

	_program.resize(AudioEffect::numPrograms);

	// copy presets into the program
	for(i = 0; i < _program.size() && i < g_num_blk_fx_presets; i++)
		_program[i] = g_blk_fx_presets[i];

	setNumInputs (1);				// mono input
	setNumOutputs(1);				// mono output

	hasVu ();
	hasClip(false);
	canProcessReplacing(true);
	programsAreChunks(true);

	// this was registered from steinberg
	setUniqueID ('g319');

	editor = new Gui(this);

	init();

	resume ();						// flush buffer
}

//-------------------------------------------------------------------------------------------------
DtBlkFx::~DtBlkFx ()
{
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::init()
// clear out all buffers
{
	// current absolute sample position of input/output updated with every input buf
	// i.e. absolute position of _x0[_x0_i+x0_n] / _x3[_x3_o]
	_curr_samp_abs = 0;
	_next_blk_fwd_n = 0;

	// absolute sample position of start of current fft blk at x0[x0_i]/x1[0]/x2[0]
	_blk_samp_abs = 0;
	_x0_n_past_end = 0;		// number of samples copied past end of buffer
	_x0_i = 0;				// pre-fft buffer offset
	_x0_n = 0;				// number of samples in the pre-fft buffer

	_x3_o = 0;				// output FIFO output index
	_x3_end_abs = 0;		// no data in _x3

	// clear out the output buffer
	memset(&_x3[0], 0, _x3.size()*sizeof(float));
	_x3_is_clear = true;

	// these will be updated on first poll
	sampleRate = 44100.0f;
	_prev_poll_abs = 0;
	_samps_per_poll = 0;
	_samps_per_tick = 250.0f;
	_samps_per_beat = 4000.0f;
	_beat_start_abs = 0;

	// cause effects to initialize
	for(int i = 0; i < BlkFxParam::NUM_FX_SETS; i++) {
		_fx_params[i].prev_type = -1;
		_fx_params[i].type = -1;
	}

	_params_state = PARAMS_CHK_SYNC;
	_params_need_processing = true;

	// these variables will be updated on first paramsChk()
	_dst_fft_abs = 0;
	_freq_fft_n = 4096;

}

//-------------------------------------------------------------------------------------------------
long DtBlkFx::getChunk (void** data, bool isPreset)
// virtual, override AudioEffect
{
	DBG("DtBlkFx::getChunk" << VAR(isPreset));

	// copy current params into chunk buffer
	for(int i = 0; i < BlkFxParam::TOTAL_NUM; i++) _chunk_data.params[i] = _params.getInput(i);

	*data = &_chunk_data;
	return sizeof(_chunk_data);
}

//-------------------------------------------------------------------------------------------------
long /*1=success*/ DtBlkFx::setChunk (void* data, long byteSize, bool isPreset)
// virtual, override AudioEffect
{
	DBG("DtBlkFx::setChunk" << VAR(byteSize) << VAR(isPreset));

	if(byteSize != sizeof(_chunk_data)) return 0; // fail

	ChunkData* chunk_data = (ChunkData*)data;
	if(chunk_data->version != ChunkData::VERS0 || chunk_data->tag != ChunkData::TAG)
		return 0; // fail

	for(int i = 0; i < BlkFxParam::TOTAL_NUM; i++) setParameter(i, chunk_data->params[i]);

	// should we save programs too??

	return 1; //ok
}


//-------------------------------------------------------------------------------------------------
void DtBlkFx::setProgram (long num)
// virtual, override AudioEffect
// called by vst to set a program
{
	AudioEffect::setProgram(limit_range(num, 0, (int)_program.size()-1));

	for(int i = 0; i < BlkFxParam::TOTAL_NUM; i++) setParameter(i, _program[curProgram].params[i]);

	DBG("DtBlkFx::setProgram="<<curProgram);
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::setProgramName (char *name)
// virtual, override AudioEffect
// called by vst to change the name of a program
{
	DBG("DtBlkFx::setProgramName");
	_program[curProgram].setName(name);
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::getProgramName (char *name)
// virtual, override AudioEffect
// called by vst to get the name of a program
{
	DBG("DtBlkFx::getProgramName");
	strcpy(name, _program[curProgram].name);
}

//-------------------------------------------------------------------------------------------------
bool DtBlkFx::getProgramNameIndexed (long category, long index, char* text)
// virtual, override AudioEffect
// called by vst to get the name of a program by category
{
	DBG("DtBlkFx::getProgramNameIndexed" << VAR(category) << VAR(index) << VAR(_program.size()));
	if(index < 0 || index >= (int)_program.size()) return false;
	strcpy(text, _program[index].name);
	return true;
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::resume ()
// virtual, override AudioEffect
// called by vst to get ready to start processing
{
	DBG("DtBlkFx::resume");
	AudioEffectX::resume ();
	if(editor) gui()->resume();
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::suspend ()
// virtual, override AudioEffect
// called by vst when sound is stopping
{
	DBG("DtBlkFx::suspend");
	AudioEffectX::suspend();

	// roll back params to just contain most recent and reset sample position
	_params.resetAndCopyIn();

	// clear buffers
	init();

	if(editor) gui()->suspend();
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::setBlockSize(long sz)
// virtual, override AudioEffect
// called by vst to indicate max number of samples that will be passed to process
{
	AudioEffectX::setBlockSize(sz);
	_max_delay_n = _x3.size()-MAX_FFT_BLK_SZ-sz;
}

struct MyInfo : public VstTimeInfo
{
	void dbgprint() {
		DBG("VstTimeInfo" 
			<< VAR(samplePos)
			<< VAR(sampleRate)
			<< VAR(nanoSeconds)
			<< VAR(ppqPos)
			<< VAR(tempo)
			<< VAR(barStartPos)
			<< VAR(cycleStartPos)
			<< VAR(cycleEndPos)
			<< VAR(timeSigNumerator)
			<< VAR(timeSigDenominator)
			<< VAR(smpteOffset)
			<< VAR(smpteFrameRate)
			<< VAR(samplesToNextClock)
			<< VAR(flags)			
		);
	}
};
//-------------------------------------------------------------------------------------------------
float/*hz*/DtBlkFx::guessRoundHz(float freq_hz, float bin_adjust_frac) const
// given a freq, round using the guessed FFT len
{
	bool neg = false;
	if(freq_hz < 0.0f) {
		freq_hz = -freq_hz;
		neg = true;
	}
	float fft_n = guessFFTLen();
	float hz_per_bin = sampleRate/fft_n;
	float bin = limit_range(freq_hz/hz_per_bin, 0.0f, fft_n/2.0f);
	float r = (floorf(bin+0.5f)+bin_adjust_frac)*hz_per_bin;
	return neg?-r:r;
}

//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::pollUpdate(bool force)
// internal method
//
// update polled variables from vst host (to do with tempo)
//
{
	if(!force && _curr_samp_abs - _prev_poll_abs < _samps_per_poll) return;

	// ask host for tempo
	VstTimeInfo* ti = getTimeInfo(kVstTempoValid|kVstPpqPosValid);
	//((MyInfo*)ti)->dbgprint();

	if(ti->flags & kVstTempoValid) 
		_samps_per_beat = (float)(60.0*sampleRate/ti->tempo);
	
	// calculate samples per tick (based on tempo)
	_samps_per_tick = (1.0f/(float)BlkFxParam::TICKS_PER_BEAT)*_samps_per_beat;

	// interpolate params 1/4 beat before the value changes
	_params.setExpectedDist((long)(_samps_per_beat*.25f));

	float ppq_pos = (float)ti->ppqPos;
	float ppq_frac = ppq_pos-floorf(ppq_pos);

	// work out sample position of the start of the current beat
	if(ti->flags & kVstPpqPosValid)
		_beat_start_abs = _curr_samp_abs-(long)(ppq_frac*_samps_per_beat);

	//DBG("pollUpdate" <<VAR(_samps_per_beat)<<VAR(ppq_pos)<<VAR(ppq_frac)<<VAR(_beat_start_abs));

	// poll every beat
	_samps_per_poll = (long)(_samps_per_beat);
	_prev_poll_abs = _curr_samp_abs;
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::setParameter (long index, float value)
// virtual, override AudioEffect
// called by vst to set param
{
	using namespace BlkFxParam;
	if(index < 0 || index >= TOTAL_NUM) return;

	if(_params.put(_curr_samp_abs, index, value) /*already exists*/ == false) {
		// this is a new time
		pollUpdate(/*force*/true);

		// force processing loop to try re-interpolating the params if it hasn't already
		if(_params_state != PARAMS_INTERP_OK)
			_params_need_processing = true;
	}

	if (editor)
		((AEffGUIEditor*)editor)->setParameter (index, value);
	//if (editor) editor->postUpdate ();
}

//-------------------------------------------------------------------------------------------------
float DtBlkFx::getParameter (long index)
// virtual, override AudioEffect
// called by vst to get param value
{
	using namespace BlkFxParam;
	if(index < 0 || index >= TOTAL_NUM) return 0.0f;
	// use most recently input value
	return _params.getInput(index);
}


//-------------------------------------------------------------------------------------------------
void DtBlkFx::getParameterName (long index, char *label)
// virtual, override AudioEffect
// called by vst to get name of param
{
	using namespace BlkFxParam;
	SplitParamNum p(index);

	// try global params
	switch(p.glob_param) {
	case MIX_BACK: strcpy(label, "Mix back"); return;
	case DELAY: strcpy(label, "Delay"); return;
	case FFT_LEN: strcpy(label, "Blk Len"); return;
	case OVERLAP: strcpy(label, "Overlap"); return;
	}
	// try fx set
	switch(p.fx_param) {
	case FX_FREQ_A: sprintf(label, "%d: Freq A", p.fx_set); return;
	case FX_FREQ_B: sprintf(label, "%d: Freq B", p.fx_set); return;
	case FX_AMP: sprintf(label, "%d: Rel Amp", p.fx_set); return;
	case FX_TYPE: sprintf(label, "%d: Effect", p.fx_set); return;
	case FX_VAL: sprintf(label, "%d: Value", p.fx_set); return;
	}
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::getParameterLabel (long index, char *label)
// virtual, override AudioEffect
// called by vst to get param label
{
	using namespace BlkFxParam;
	SplitParamNum p(index);
	switch(p.glob_param) {
	case MIX_BACK: strcpy(label, "%"); return;
	case DELAY: strcpy(label, "beats"); return;
	case FFT_LEN: strcpy(label, "sec"); return;
	case OVERLAP: strcpy(label, "%"); return;
	}
	switch(p.fx_param) {
	case FX_FREQ_A: strcpy(label, "Hz"); return;
	case FX_FREQ_B: strcpy(label, "Hz");  return;
	case FX_AMP: strcpy(label, "dB");  return;
	case FX_TYPE: strcpy(label, "");  return;
	case FX_VAL: strcpy(label, "");  return;
	}
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::getParameterDisplay (long index, char *text)
// virtual, override AudioEffect
// called by vst to convert a param to text
{
	using namespace BlkFxParam;
	SplitParamNum p(index);

	if(!p.ok) {
		sprintf(text, "?");
		return;
	}

	// get current input
	float v = _params.getInput(index);

	// see if index belongs to the global params
	switch(p.glob_param) {
	case MIX_BACK:
		sprintf(text, "%2.0f", v*100.0f); return;

	case DELAY:
		sprintf(text, "%1.3f", getDelaySamps(v)/_samps_per_beat); return;

	case FFT_LEN: {
			float fft_len = guessFFTLen();

			// determine blk size in seconds
			float sec = fft_len/sampleRate;

			// print blk size in seconds or msec
			sprintNum(text, sec);
			return;
		}

	case OVERLAP: {
			float fft_len = guessFFTLen();
			float samps_fwd = (float)getBlkShiftFwd(v, fft_len);
			float overlap = 1.0f-(samps_fwd/fft_len);
			sprintf(text, "%s%.0f", getBlkSync(v)?"sync ":"", overlap*100.0f); return;
		}
	}

	// get the fx interface
	int fx_idx = getEffectType(
		_params.getInput( paramOffs(p.fx_set)+FX_TYPE )
	);
	FFTFx* fft_fx = GetFFTFx(fx_idx);

	switch(p.fx_param) {
	case FX_FREQ_A:
	case FX_FREQ_B: {
			float freq_hz = guessRoundHz(BlkFxParam::getHz(v));
			if(!fft_fx->paramUsed(p.fx_param)) 
				sprintf(text, "not used");
			else {
				// put in note based on freq
				text += HzToNote(text, freq_hz);
				*text++ = '/';
				// print in Hz or kHz
				sprintNum(text, freq_hz, /*min unit*/1.0f);
			}
			return;
		}
	case FX_AMP: {
			bool mix_mode = fft_fx->ampMixMode();
			float amp_mult = getEffectAmpMult(v, mix_mode);
			if(!fft_fx->paramUsed(FX_AMP))
				sprintf(text, "not used");
			else if(mix_mode && amp_mult <= 1.0f)
				sprintf(text, "%.0f %%", amp_mult*100.0f);
			else if(amp_mult > 0)
				sprintNum(text, getEffectAmp(v), /*min unit*/1.0f, /*max unit*/100.0f);
			else
				sprintf(text, "-inf");
			return;
		}

	case FX_TYPE:
		strcpy(text, GetFFTFx(getEffectType(v))->name());
		return;

	case FX_VAL:
		GetFFTFx(fx_idx)->dispVal(
			this,		//
			text,		// output
			v,			// value
			p.fx_set	// which fx set we are
		);
		return;
	}
}


//-------------------------------------------------------------------------------------------------
bool DtBlkFx::getEffectName (char* name)
// virtual, override AudioEffect
// called by vst to return name of effect
{
	DBG("DtBlkFx::getEffectName");
	strcpy (name, "DTBlkFx");
	return true;
}

//-------------------------------------------------------------------------------------------------
bool DtBlkFx::getProductString (char* text)
// virtual, override AudioEffect
// called by vst to get product string
{
	DBG("DtBlkFx::getProductString");
	strcpy (text, "DtBlkFx");
	return true;
}

//-------------------------------------------------------------------------------------------------
bool DtBlkFx::getVendorString (char* text)
// virtual, override AudioEffect
// called by vst to get vendor string
{
	DBG("DtBlkFx::getVendorString");
	strcpy (text, "DT");
	return true;
}


//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::doFFT()
// internal method
//
// perform FFT, find accumulative power of spectrum & scale to prevent overflow when
// doing effects
//
{
	// do the FFT
	rfftw_one(g_plan_fft[_plan], &_x0[_x0_i], &_x1[0]);

	// scale spectrum
	float scale = 1.0f/_freq_fft_n;

	// set imaginary part of dc to be 0 (so 0hz is now a complex number)
	_x1[_freq_fft_n] = 0;


	// find power of spectrum
	float acc = 0.0f; 
	pcplxf_range in(FFTdata(), /*start*/0, /*end*/_freq_fft_n/2);
	for(; in.r != in.e; in++) {
		*in = *in * scale;
		acc += norm(*in);
	}

	// nyquist freq is real only
	*(in.r) *= scale;
	acc += *(in.r) * *(in.r); 

	_total_out_pwr = acc;
	_total_in_pwr = acc;

	if(editor)
		gui()->_sgram[0].newData(_blk_samp_abs, _time_fft_n, _plan, /*pwr scale*/1.0f);
}


//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::doIFFT()
// internal method
// inverse FFT freq back to time domain
{
	// set to undo scaling & match input power
	double pwr_scale = 1.0f;
	if(_total_out_pwr > 1e-30) pwr_scale *= _total_in_pwr/_total_out_pwr;
	if(pwr_scale > 1e30) pwr_scale = 1.0f;			// too big
	_x2_scale = (float)sqrt(pwr_scale);

	//DBG("ifft" << VAR(pwr_scale) << VAR(_total_in_pwr) << VAR(_total_out_pwr) << VAR(_x2_scale));
	if(editor)
		gui()->_sgram[1].newData(_blk_samp_abs, _time_fft_n, _plan, (float)pwr_scale);

	// perform inverse FFT
	rfftw_one(g_plan_ifft[_plan], &_x1[0], &_x2[0]);
}

//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::processFFT()
// internal method
// process the FFT'd data (x1) using the fx_params
//
{
	float fft_len = (float)_freq_fft_n;

	// collect all parameters first
	for(_fx_set = 0; _fx_set < BlkFxParam::NUM_FX_SETS; _fx_set++) {
		FxParams& p = currFxParams();

		int param_offs = BlkFxParam::paramOffs(_fx_set);
		p.freq_param[0] = _params.getInterp(BlkFxParam::FX_FREQ_A+param_offs);
		p.freq_param[1] = _params.getInterp(BlkFxParam::FX_FREQ_B+param_offs);
		p.amp_param = _params.getInterp(BlkFxParam::FX_AMP+param_offs);
		p.val = _params.getInterp(BlkFxParam::FX_VAL+param_offs);
		long type = BlkFxParam::getEffectType(
			_params.getNonInterp(BlkFxParam::FX_TYPE+param_offs)
		);

		p.prev_type = p.type;
		p.type = type;
		p.fft_fx = GetFFTFx(type);

		// convert the params
		p.fbin[0] = BlkFxParam::getFFTBin(p.freq_param[0], fft_len, sampleRate);
		p.fbin[1] = BlkFxParam::getFFTBin(p.freq_param[1], fft_len, sampleRate);
		p.amp = BlkFxParam::getEffectAmpMult(p.amp_param, p.fft_fx->ampMixMode());

		// round bins to ints
		p.bin[0] = (long)(p.fbin[0]+0.5f);
		p.bin[1] = (long)(p.fbin[1]+0.5f);
	}

	// run each effect
	for(_fx_set = 0; _fx_set < BlkFxParam::NUM_FX_SETS; _fx_set++)
		currFxParams().fft_fx->process(this);

	// zero the spectrum extremes
	_x1[0] = 0;
	_x1[_freq_fft_n/2] = 0;
}


//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::mixToOutputBuffer(float mix_back)
// internal method
// mix from x2 (post-IFFT) & x0 (pre-FFT) into x3
// x2*(1-mix_back) + x0*mix_back => x3
//
{
	// the amount of xfade-in corresponds to the overlap of this fft blk with existing data
	long fadein_n = _x3_end_abs-_dst_fft_abs;

	// normal case is not to xfade-out current fft-blk because next blk will do the xfade
	long fadeout_n = 0;

	// get the absolute ending position of blk
	long dst_fft_end_abs = _dst_fft_abs+_time_fft_n;

	// how much the new fft blk will increase the amount of data in x3
	long extend_x3 = dst_fft_end_abs - _x3_end_abs;

	if(fadein_n >= 0) {
		// normal case: previous data and the current fft blk overlap, cross-fade

		if(extend_x3 < 0) {
			// special case: current fft blk ends up entirely within existing x3 data, 

			// do nothing if the blk ends prior to the current output position (should not happen)
			if(_curr_samp_abs - dst_fft_end_abs >= 0) {
				DBG("mixToOutputBuffer possible ERROR" << VAR(_curr_samp_abs - dst_fft_end_abs));
				return;
			}

			// cross fade all of fft blk in & out
			fadein_n = _time_fft_n/2;
			fadeout_n = _time_fft_n-fadein_n;

			// one case is not handled, part of current blk is before current sample position
			// (should never happen)
		}
	}
	else {
		// special case: there is a gap between the end of prev data and the current fft blk

		// zero-fill gap (and a bit extra because we will still cross-fade to the current blk)
		long zero_o = _x3_end_abs-_curr_samp_abs;
		long zero_n = -fadein_n;

		// arbitrary fade in of current fft blk
		fadein_n = 16;

		// zero data in the gap (unless x3 is already cleared)
		if(!_x3_is_clear) wrapProcess(PZero(), _x3, _x3_o+zero_o, zero_n+fadein_n);
		_x3_is_clear = false;

		// arbitrary fade out of previous data 
		long prev_fade_n = 16;
		long prev_fade_o = zero_o-prev_fade_n;

		// don't fade data that has already been output
		if(prev_fade_o < 0) {
			prev_fade_n += prev_fade_o;
			prev_fade_o = 0;
		}
		if(prev_fade_n > 0)
			wrapProcess(PFadeOut(prev_fade_n), _x3, _x3_o+prev_fade_o, prev_fade_n);

		//DBG("mixToOutputBuffer gap" << VAR(zero_o) << VAR(zero_n) << VAR(prev_fade_n) << VAR(prev_fade_o));
	}

	// ensure fade in is not too big
	if(fadein_n > _time_fft_n) fadein_n = _time_fft_n;

	// offset into x3 at which to cross fade in first sample of fft blk
	long x3_fadein_o = _dst_fft_abs-_curr_samp_abs+_x3_o;

	// offset into x3 at which to start copy of curr fft blk
	long x3_cp_o = x3_fadein_o + fadein_n;

	// number of samples to copy from fft blk
	long cp_n = _time_fft_n - fadein_n - fadeout_n;

	// offset into x3 at which to start fade-out
	long x3_fadeout_o = x3_cp_o+cp_n;

	/*DBG("mixToOutputBuffer"
		<< VAR(fadein_n)
		<< VAR(fadeout_n)
		<< VAR(extend_x3)
		<< VAR(x3_fadein_o)
		<< VAR(x3_cp_o)
		<< VAR(cp_n)
		<< VAR(x3_fadeout_o)
		<< VAR(mix_back)
	);*/

	// do fade-in, direct copy, fade-out
	if(mix_back <= 0.0f) {
		// output data is completely fft blk
		if(fadein_n)
			wrapProcess(PXFade(&_x2[0], _x2_scale, fadein_n), _x3, x3_fadein_o, fadein_n);
		if(cp_n)
			wrapProcess(PScaleCopy(&_x2[fadein_n], _x2_scale), _x3, x3_cp_o, cp_n);
		if(fadeout_n)
			wrapProcess(PXFade(&_x2[fadein_n+cp_n], _x2_scale, fadeout_n, /*rev*/true),
				_x3, x3_fadeout_o, fadeout_n);

		//wrapProcess(PScaleCopy(&_x2[0], _x2_scale), _x3, x3_fadein_o, _time_fft_n);

	}
	else if(mix_back >= 1.0f) {
		// output data is completely original
		if(fadein_n)
			wrapProcess(PXFade(&_x0[_x0_i], 1.0f, fadein_n), _x3, x3_fadein_o, fadein_n);
		if(cp_n)
			wrapProcess(PScaleCopy(&_x0[_x0_i+fadein_n], 1.0f), _x3, x3_cp_o, cp_n);
		if(fadeout_n)
			wrapProcess(PXFade(&_x0[_x0_i+fadein_n+cp_n], 1.0f, fadeout_n, /*rev*/true),
				_x3, x3_fadeout_o, fadeout_n);
	}
	else {
		// output data is a mix of original and processed
		_x2_scale *= 1.0f-mix_back; // processed data
		float x0_scale = mix_back; // original data

		if(fadein_n)
			wrapProcess(PXFade2(&_x2[0], _x2_scale, &_x0[_x0_i], x0_scale, fadein_n),
				_x3, x3_fadein_o, fadein_n);
		if(cp_n)
			wrapProcess(PScaleCopy2(&_x2[fadein_n], _x2_scale, &_x0[_x0_i+fadein_n], x0_scale),
				_x3, x3_cp_o, cp_n);
		if(fadeout_n)
			wrapProcess(PXFade2(&_x2[fadein_n+cp_n], _x2_scale, &_x0[_x0_i], x0_scale, fadeout_n, /*rev*/true),
				_x3, x3_fadeout_o, fadeout_n);
	}

	// update how much data has been put into x3
	if(extend_x3 > 0)
		_x3_end_abs = _dst_fft_abs+_time_fft_n;
}

//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::nextBlk()
// internal method
// update everything ready for the next block
//
{
	float overlap_param = _params.getInterp(BlkFxParam::OVERLAP);

	_next_blk_fwd_n = BlkFxParam::getBlkShiftFwd(overlap_param, (float)_time_fft_n);

	// synchronize next blk to start-of-beat position if blk_sync is true
	if(BlkFxParam::getBlkSync(overlap_param)) {
		// next blk sample position as determined from current overlap param
		long next_blk_samp_abs = _blk_samp_abs+_next_blk_fwd_n;
				
		// find the sample position of the most recent beat before next blk position
		float beats = _samps_per_beat>0.0f ?
			ceilf((_beat_start_abs-next_blk_samp_abs)/_samps_per_beat)
			: 1/*this case should never happen*/;

		long sync_abs = (long)(_beat_start_abs-beats*_samps_per_beat);
		long sync_fwd = sync_abs-_blk_samp_abs;

		//DBG("nextBlk, blksync" <<VAR(_next_blk_fwd_n) <<VAR(sync_fwd) <<VAR(_blk_samp_abs) <<VAR(next_blk_samp_abs)<<VAR(beats)<<VAR(sync_abs)<<VAR(overlap_param));
		if(sync_fwd > max(100L, (long)_samps_per_tick) && sync_fwd <= _next_blk_fwd_n) _next_blk_fwd_n = sync_fwd;
	}

	_params_state = PARAMS_CHK_SYNC;
	_params_need_processing = true;
}


//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::paramsChk()
// internal method
//
// check whether params need processing & update variables used in processing loop
//
{
	if(!_params_need_processing) return;
	_params_need_processing = false;

	// chk if the next param is available and possibly sync to it
	if(_params_state == PARAMS_CHK_SYNC && _params.setOutPos(_blk_samp_abs+1, /*find fraction*/false)) {

		// should we sync fft blk with param?
		if(BlkFxParam::getBlkSync(_params.getNonInterp(BlkFxParam::OVERLAP, /*next param*/true)) && 
			_params.isAnyExplicitSet(/*next param*/true)) {

			// find abs sample position of param
			long param_abs = _params.getSampAbs(/*next param*/true);

			// how far ahead of the current position is this?
			long param_fwd = param_abs-_blk_samp_abs;

			//DBG("paramsChk, blksync" <<VAR(_params.getSampAbs(false)) <<VAR(param_abs) <<VAR(param_fwd) <<VAR(_next_blk_fwd_n));
			if(param_fwd > max(100L, (long)_samps_per_tick) && param_fwd < _next_blk_fwd_n) _next_blk_fwd_n = param_fwd;
		}
		_params_state = PARAMS_NONINTERP;
	}

	// find src position of fft blk
	long src_fft_abs = _blk_samp_abs + _next_blk_fwd_n;

	if(_params_state == PARAMS_NONINTERP) {
		// see if we can interpolate
		if(_params.setOutPos(src_fft_abs))
			_params_state = PARAMS_INTERP_OK;
	}

	// convert params
	//
	_delay_n = getDelaySamps(_params.getInterp(BlkFxParam::DELAY));
	_plan = BlkFxParam::getPlan(_params.getInterp(BlkFxParam::FFT_LEN));

	// calculate derived variables
	//

	// find output position of the current blk using current delay
	_dst_fft_abs = src_fft_abs + _delay_n;

	// make sure dest position isn't behind current sample position
	long behind_n = _curr_samp_abs-_dst_fft_abs;
	if(behind_n > 0) {
		_next_blk_fwd_n += behind_n;
		_dst_fft_abs = _curr_samp_abs;
	}

	// number of samples to do fft blk
	_freq_fft_n = GET_FFT_BLK_SZ(_plan);
}

//-------------------------------------------------------------------------------------------------
inline void DtBlkFx::doInBuf(float* in_buf, long buf_n)
// internal method
// copy input data into the pre-fft buffer
{
	// offset into buffer at which to write data
	long i = _x0_i+_x0_n; 

	// wrap to the start of the buffer if need be
	if(i >= _x0_sz) i -= _x0_sz;

	// check if copy will pass _x0_sz
	else if(i+buf_n > _x0_sz) {
		// copy data that will overflow past _x0_sz to the start of the buffer
		long skip = _x0_sz - i;
		long n = buf_n - skip;
		memcpy(&_x0[0], in_buf+skip, n*sizeof(float));

		// number of samples past _x0_sz
		_x0_n_past_end = n;
	}

	// copy all data into buffer (note that this may be past _x0_sz)
	memcpy(&_x0[i], in_buf, buf_n*sizeof(float));
	_x0_n += buf_n;
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::_process(float* in_buf, long buf_n)
// internal method
{
	//DBG("_process" << VAR(buf_n));
		
	pollUpdate(/*force*/false);

	doInBuf(in_buf, buf_n);

	// 1 + absolute position of the final sample to output now
	long buf_end_abs = _curr_samp_abs+buf_n;

	// process fft-blks in a loop
	while(_x0_n) {
		paramsChk();

		// determine whether we are forced to output a blk because of the delay or
		// we've run out of input buffer
		bool force_out = (buf_end_abs-_dst_fft_abs > 0 || _x0_n>=_x0_force_out_sz);

		// true if there's enough data in x0 to do the fft
		bool enough_data = (_x0_n >= _next_blk_fwd_n+_freq_fft_n);

		/*DBG("_process loop"
			<< VAR(_curr_samp_abs)
			<< VAR(_dst_fft_abs)
			<< VAR(_blk_samp_abs)
			<< VAR(_x0_i)
			<< VAR(_x0_n)
			<< VAR(_x0_n_past_end)
			<< VAR(_x3_o)
			<< VAR(_x3_end_abs)
			<< VAR(_params_state)
			<< VAR(_delay_n)
			<< VAR(_freq_fft_n)
			<< VAR(force_out)
			<< VAR(enough_data)
		);*/

		// stop looping if we're not forced to output and either the params aren't ready
		// or we don't have enough data for a blk
		if(!force_out && (_params_state != PARAMS_INTERP_OK || !enough_data)) break;

		// move forward in the input buffer
		_blk_samp_abs += _next_blk_fwd_n;
		if(_x0_n >= _next_blk_fwd_n) {
			// normal case 
			_x0_i += _next_blk_fwd_n;
			_x0_n -= _next_blk_fwd_n;

			// see whether we need to wrap to the start of the buffer (assume fwd_n is < _x0_sz)
			if(_x0_i >= _x0_sz) {
				_x0_i -= _x0_sz;
				_x0_n_past_end = 0; // no samples have been copied past the end of the buffer
			}
		}
		else {
			// special case: should never happen - all data in _x0 is used up, reset
			_x0_i = 0;
			_x0_n = 0;
			_x0_n_past_end = 0;
		}

		// process the data blk
		
		// check whether there's enough time domain data
		_time_fft_n = _freq_fft_n;
		if(!enough_data) {
			// delay is too short for requested blk sz, reduce fft length
			_plan = reducePlan(_plan, _x0_n);
			_freq_fft_n = GET_FFT_BLK_SZ(_plan);

			// zero-pad input data
			long zero_n = _freq_fft_n-_x0_n;
			float* x0 = &_x0[_x0_i+_x0_n];
			while(zero_n-- > 0) *x0 = 0.0f; 

			// reduce time domain fft len to match the amount of data we have
			_time_fft_n = _x0_n;
		}


		// make a contiguous blk if the current blk is split at the end of the FIFO
		long x0_sz_ext = _x0_sz+_x0_n_past_end;
		long split_n = (_x0_i+_time_fft_n)-x0_sz_ext;
		if(split_n > 0) {
			memcpy(&_x0[x0_sz_ext], &_x0[_x0_n_past_end], split_n*sizeof(float));
			_x0_n_past_end += split_n;
		}

		// fraction of original data
		float mix_back = _params.getInterp(BlkFxParam::MIX_BACK);
		if(mix_back < 1.0f) {
			doFFT();
			processFFT();
			doIFFT();
		}
		mixToOutputBuffer(mix_back);
		nextBlk();
	}

	// zero fill x3 if it has no data to output for this blk - this will happen before the first
	// fft blk has been processed (although x3 will already be cleared) or if the delay is 
	// increased in some cases
	//
	long zero_n = buf_end_abs-_x3_end_abs;
	if(zero_n > 0) {
		long zero_o = _x3_end_abs-_curr_samp_abs;
		wrapProcess(PZero(), _x3, _x3_o+zero_o, zero_n);

		// fade out previous data (if there's any in this blk)
		long fade_n = 16;
		long fade_o = zero_o-fade_n;
		if(fade_o < 0) {
			fade_n = fade_o;
			fade_o = 0;
		}
		if(fade_n > 0)
			wrapProcess(PFadeOut(fade_n), _x3, _x3_o+fade_o, fade_n);

		_x3_end_abs = buf_end_abs;
	}

	// update absolute sample position
	_curr_samp_abs += buf_n;
}



//-------------------------------------------------------------------------------------------------
void DtBlkFx::process (float** inputs, float** outputs, long samps)
// virtual, override AudioEffect
// called by vst to process data from "inputs" and add to "outputs"
{
	_process(inputs[0], samps);
	wrapProcess(PAddOut(outputs[0]), _x3, _x3_o, samps);
	_x3_o = wrap(samps+_x3_o, _x3);
}

//-------------------------------------------------------------------------------------------------
void DtBlkFx::processReplacing (float** inputs, float** outputs, long samps)
// virtual, override AudioEffect
// called by vst to process data from "inputs" and replace data in "outputs"
{
	_process(inputs[0], samps);
	wrapProcess(PCopyOut(outputs[0]), _x3, _x3_o, samps);
	_x3_o = wrap(samps+_x3_o, _x3);
}




