#ifndef _DT_BLK_FX_H_
#define _DT_BLK_FX_H_
/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

Effect main class

History
	Date       Version    Programmer         Comments
	16/2/03    1.0        Darrell Tam		Created


This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/

#include "audioeffectx.h"
#include <vector>
#include <dt/cplx.h>

#include <dt/windows_support.h>
#include <dt/misc_stuff.h>
#include <dt/ParamsDelay.h>
#include "BlkFxParam.h"
#include "BlkFxProgram.h"

// gui associated with DtBlkFx
class Gui;
class FFTFx;

//------------------------------------------------------------------------
class DtBlkFx : public AudioEffectX
{
public:

	DtBlkFx (audioMasterCallback audioMaster);
	~DtBlkFx ();

public: // override AudioEffectX methods
	virtual void process (float **inputs, float **outputs, long sampleframes);
	virtual void processReplacing (float **inputs, float **outputs, long sampleFrames);
	virtual void setBlockSize(long samps);

	virtual void setProgram (long program);
	virtual void setProgramName (char *name);
	virtual void getProgramName (char *name);
	virtual bool getProgramNameIndexed (long category, long index, char* text);

	virtual long getChunk (void** data, bool isPreset = false);
	virtual long setChunk (void* data, long byteSize, bool isPreset = false);

	virtual void setParameter (long index, float value);
	virtual float getParameter (long index);
	virtual void getParameterLabel (long index, char *label);
	virtual void getParameterDisplay (long index, char *text);
	virtual void getParameterName (long index, char *text);

	virtual void resume();
	virtual void suspend();

	virtual bool getEffectName (char* name);
	virtual bool getVendorString (char* text);
	virtual bool getProductString (char* text);
	virtual long getVendorVersion () { return 1000; }
	
	virtual VstPlugCategory getPlugCategory () { return kPlugCategEffect; }

public: // our own methods
	Gui* gui() const { return (Gui*)editor; }

	// return fixed-position fft buffer
	//float* getFFTdat() { return &_x1[0]; }
	pcplxf FFTdata() { return pcplxf(&_x1[0], _freq_fft_n); }

	// reduce fft plan to be just big for "avail_samps"
	static long reducePlan(long plan/*start*/, long avail_samps);

	// guess what the current FFT length will be (for display purposes)
	float/*samples*/guessFFTLen() const;

	// guess what a frequency will be rounded to (for display purposes)
	float/*hz*/guessRoundHz(float freq_hz, float bin_adjust_frac=0.0f) const;

	long/*samples*/getDelaySamps(float/*0..1*/delay_param) const;

	float getSampsPerBeat() const { return _samps_per_beat; }

	// get the most recently set param 
	float getCurrParam(int idx) const { return _params.getInput(idx); }

protected: // internal methods
	void init();
	void doFFT();
	void doIFFT();
	void processFFT();
	void mixToOutputBuffer(float mix_back);
	void nextBlk();
	void doInBuf(float* in_buf, long buf_n);
	void paramsChk();
	void _process(float* in_buf, long buf_n);
	
public: //
	// maximum number of samples delay
	long _max_delay_n;

	// current absolute sample position of input/output updated with every input buf
	// i.e. absolute position of _x0[_x0_i+x0_n] / _x3[_x3_o]
	long _curr_samp_abs;

	// number of samples to move forward in input buffer to process next blk
	long _next_blk_fwd_n;

	// absolute sample position of x0[x0_i] and also x1[0]/x2[0] when processing a fft-blk
	long _blk_samp_abs;

	std::vector<float> _x0;	// pre FFT circular buffer
	long _x0_i;				// first sample of current blk in _x0
	long _x0_n;				// number of samples in the pre-fft buffer
	long _x0_sz;			// wraping position of x0
	long _x0_force_out_sz;	// force output if x0_n exceeds this
	long _x0_n_past_end;	// number of samples copied past _x0_sz

	std::vector<float> _x3;	// output FIFO
	long _x3_o;				// output FIFO output index
	long _x3_end_abs;		// 1 + abs sample position of final sample in _x3
	bool _x3_is_clear;		// true when x3 is initialized with 0's

	enum {
		PARAMS_CHK_SYNC,	// params need to be processed for the current blk
		PARAMS_NONINTERP,	// params have been gathered but can't be interpolated
		PARAMS_INTERP_OK	// params can be properly interpolated for the current blk
	} _params_state;
	bool _params_need_processing;

	long _plan;				// actual plan (different from desired plan if forced to output data early)
	long _delay_n;			// output delay
	long _freq_fft_n;		// actual fft blk sz processed (may not correspond to _desired_plan if output forced early)


public: // temporary variables used once an fft blk is being processed
	// destination absolute sample position for current fft blk
	long _dst_fft_abs;

	long _time_fft_n;		// time domain fft blk sz (may not be pwr of 2 if forced to output data early)

	std::vector<float> _x1;	// FFT'd data (frequency-domain), scaled 1e-3
	std::vector<float> _x2;	// IFFT'd data (time-domain)

	float _x2_scale;	// scaling to apply to x2 during mix back
	float _total_in_pwr;// x1 input power
	float _total_out_pwr;// current output power

	
public: // temporary variables available for FFTFx::process() methods
	int _fx_set;		// which effect set we're currently running
	class FxParams
	{
	public:
		// unprocessed freq param, 0..1
		float freq_param[2];

		// unprocessed amp param, 0..1
		float amp_param;

		// which effect we're currently running
		FFTFx* fft_fx;
		long type;

		long prev_type;	// previous effect

		float fbin[2];	// floating point version of start & stop bins
		float amp;		// current amplitude (converted to linear scaling fraction)
		float val;		// current effect value

		long bin[2];	// start & stop bin for this effect

		// effects can store things here, this data is maintained between blocks
		std::vector<char> data;

		// true if the effect has changed this block
		bool changed() { return type != prev_type; }

	protected:
		friend DtBlkFx;
		FxParams() {}

	private:
		FxParams(const& FxParams) {} // don't allow accidental copy
	};
	FxParams _fx_params[BlkFxParam::NUM_FX_SETS];

public:	// only call these from effect processing since they rely on "_fx_set" being valid
	FxParams& currFxParams() { return _fx_params[_fx_set]; }

public: // polled variables that are updated periodically
	// samples per beat
	float _samps_per_beat;

	// there are "TICKS_PER_BEAT" ticks per beat
	float _samps_per_tick;

	// most recent start of beat sample position
	long _beat_start_abs;

	// last sample position bpm was polled
	long _prev_poll_abs;

	// number of samples before we do 
	long _samps_per_poll;

	// check whether enough samples have elapsed to update polled variables
	void pollUpdate(bool force);

protected:
	// use params delay to store trail of params
	ParamsDelay<BlkFxParam::TOTAL_NUM> _params;

	// parameter storage for getChunk()
	struct ChunkData {
		enum {
			VERS0 = 0,
			TAG = 0x99887766
		};
		unsigned long tag;
		unsigned long version;
		float params[BlkFxParam::TOTAL_NUM];
	};
	ChunkData _chunk_data;

	// array of programs
	std::vector<BlkFxProgram> _program;
};


//-------------------------------------------------------------------------------------------------
inline long DtBlkFx::getDelaySamps(float/*0..1*/delay_param) const
// internal method
// given the vst param, return the corresponding number of samples delay
{
	return limit_range(
		(long)(BlkFxParam::getDelayTicks(delay_param)*_samps_per_tick),
		FFT_BLK_SZ_0*2,		// arbitary minimum size
		_max_delay_n		// max delay
	);
}
//-------------------------------------------------------------------------------------------------
inline long DtBlkFx::reducePlan(long plan/*start*/, long avail_samps)
// static method
// find smallest fft plan that can fit "avail_samps" starting at the given plan
//
{
	while(plan > 0 && GET_FFT_BLK_SZ(plan-1) > avail_samps) plan--;
	return plan;
}

//-------------------------------------------------------------------------------------------------
inline float/*samples*/DtBlkFx::guessFFTLen() const
//
// guess the FFT length that will be used from the most recently buffered params for display
// params - this may not be the actual length because the input buffer may be full and it
// depends on the exact process buffer
//
{
	long delay_n = getDelaySamps(_params.getInput(BlkFxParam::DELAY));
	return (float)GET_FFT_BLK_SZ(
		reducePlan(
			BlkFxParam::getPlan(
				_params.getInput(BlkFxParam::FFT_LEN)
			),
			delay_n
		)
	);
}


#endif
