/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

Implementation of FFTFx classes - these are the actual FFT effects called by DtBlkFx

There is one FFTFx class instance per effect type

In general the call sequence for processing data is as follows:
1. DtBLkFx has buffered a blk of data, performed FFT
2. DtBlkFx calls GetFFTFx function to retrieve the specific effect instance derived from FFTFx class
3. DtBlkFx calls the "process" method on instance from (2)
4. Process method instantiates an Effect processing class (on the stack) and calls the "Run" template function
   with the appropriate Effect
5. The run template (usually) instantiates a SplitMaskProcess (which handles processing the range
   between freqA & freqB or excluding the range between freqA & freqB)
6. The run template may instantiate another MaskProcess if the previous effect is say a harmonic mask
7. The run method from the SplitMaskProcess (or other MaskProcess) is called
8. SplitMaskProcess (or other MaskProcess) calls run method of effect class passed into Run template
   function 0 or more times for each continuous piece of the spectrum that needs processing
   
The key things are:
   Heavy use of templates to try to force as much compile time optimization as possible.

   There are 2 levels of classes for each effect: the interface based on FFTFx where there is one instance
   per effect and the actual effect processing that is derived from the internal class "ProcessBase".

   Masks and effects are both based on "ProcessBase" and calls flow downwards via the "run" method of
   "ProcessBase" that they override.

   The ProcessBase effects are instantiated on the stack, so that means you can't store any data between
   blocks in them

History
	Date       Version    Programmer         Comments
	16/2/03    1.0        Darrell Tam		Created


This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/

//#define _DBG
#define DBG_FILE_NAME "c:\\fftfx.txt"
#include "Debug.h"

#include <dt/SinCosTable.h>

#include "process_loops.h"
#include "DtBlkFx.hpp"
#include "FFTFx.h"
#include "HarmData.h"

using namespace std;

static SinCosTable<cplxf, /*bits*/12> sincos_table;
long g_rand_i = 1;

//*************************************************************************************************
class ProcessBase
// base class for effects
{
public:
	bool reverse() { return false; }

	void prepare(ProcessBase* parent) {}
	void run(long b0, long b1) {}
	void done() {}

	float pwrChg() { return 0.0f; }
};

//*************************************************************************************************
class AmpProcess : public ProcessBase
//
// default process multiplies data by amplitude
//
{
public:
	Ptr<DtBlkFx> _b;
	pcplxf _x1;

	// amplify to apply
	float _amp;

	// input pwr
	float _in_pwr;

	void init(DtBlkFx* b, float amp)
	{
		_b = b;
		_amp = amp;
		_x1 = b->FFTdata();
		_in_pwr = 0.0f;
	}

	AmpProcess(DtBlkFx* b, float amp) { init(b, amp); }
	AmpProcess(DtBlkFx* b) { init(b, b->currFxParams().amp); }

	// apply scaling to bins b0 to b1
	void run(long b0, long b1)
	{
		for(pcplxf_range x(_x1, b0, b1+1); !x.empty(); x++) {
			_in_pwr += norm(*x);
			*x = (*x) * _amp;
		}
	}

	float pwrChg() { return (_amp*_amp-1)*_in_pwr; }
};

//*************************************************************************************************
float/*pwr*/GetPwr(pcplxf_range x/*must be fwd*/)
{
	float pwr = 0.0f;
	for(; !x.empty(); x++) pwr += norm(*x);
	return pwr;
}

float/*amp*/MatchPwr(float amp, double target_pwr, double curr_pwr)
{
	// find scaling to match output pwr to input pwr
	double s = amp;
	if(target_pwr > 0 && curr_pwr > 0) s *= sqrt(target_pwr/curr_pwr);

	// limit the scaling
	if(s > 1e30) return 1.0f;
	if(s < 1e-30) return 0.0f;
	return (float)s;

}
void MatchPwr(float amp, double target_pwr, double curr_pwr, pcplxf_range x /*must be fwd*/)
{
	amp = MatchPwr(amp, target_pwr, curr_pwr);
	for(; !x.empty(); x++)
		*x = (*x) * amp;
}

class MatchPwrProcess : public AmpProcess
// modified AmpProcess
{
public:
	// initialize amp process using amp from curr fx params
	MatchPwrProcess(DtBlkFx* b, float in_pwr, float out_pwr) : AmpProcess(b)
	{
		/*AmpProcess::*/_amp = MatchPwr(/*AmpProcess::*/_amp, in_pwr, out_pwr);
		/*AmpProcess::*/_in_pwr = in_pwr;
	}

	void run (long b0, long b1)
	{
		for(pcplxf_range x(_x1, b0, b1+1); !x.empty(); x++)
			*x = (*x) * _amp;
	}
};

//*************************************************************************************************
template <class T> class MaskProcessBase : public ProcessBase
// base class for mask processing
{
public:
	Ptr<T> _process;

	// default to calling through to next processing
	bool reverse() { return _process->reverse(); }

	void prepare(ProcessBase* parent) { _process->prepare(this); }

	void done() { _process->done(); }

	float pwrChg() { return _process->pwrChg(); }
};

//*************************************************************************************************
template <class T> class HarmMaskProcess : public MaskProcessBase<T>
{
public:
	// centre of harm-0 (bin)
	float _offs;

	// width of harmonics (bins)
	long _width;
	long _half_width;

	// spacing of harmonics (bins)
	float _spacing;

	// maximum bin (set to limit the number of harmonics processed)
	long _max_bin;

	// minimum bin (edge of lowest harmonic to process)
	long _min_bin;

	// current centre bin
	float _curr_cent;

	// current harmonic (0 = fundamental*multiplier)
	float _curr_harm;

	// original frequency supplied
	float _freq;

	// multiplier being used on fundamental
	float _multiply;

	void init(
		T& process,
		float/*0..1*/fx_val,	//
		float freq,				// bin pos of fundamental freq
		float multiply			// multiplier being applied
	)
	// initialize 
	{
		_multiply = multiply;
		/*MaskProcessBase::*/_process = &process;

		if(freq < 1) {
			// disable effect if frequency is invalid
			_multiply = 1;
			_freq = 1;
			_offs = 1;
			_spacing = 1;
			_width = 0;
			_half_width = 0;
			_max_bin = 0;
			_min_bin = 1;
			return;
		}

		_freq = freq;

		SplitParam<4> h(fx_val);

		// note, fundamental freq is considered to be numbered harmonic-1 for odd & even
		float f = freq*multiply;
		switch(h.i_part) {
		case 0: _offs = f; _spacing = f; break; // all harmonics
		case 1: _offs = f; _spacing = f*2; break; // odd harmonics
		case 2: _offs = f*2; _spacing = f*2; break; // even harmonics
		case 3: _offs = f*0.5f; _spacing = f; break; // in between harmonics
		}

		// don't let harmonics get too small
		if(_spacing < 4) _spacing = 4;

		float fwidth = limit_range(_spacing*h.f_part, 0.0f, _spacing);
		_half_width = (long)fwidth/2;
		_width = _half_width*2;

		// don't do more than 100 harmonics
		_max_bin = (long)(_spacing*100);

		// don't go below fundamental (i.e. don't treat 0Hz as a harmonic)
		_min_bin = (long)_offs-_half_width;

	}

	// if you don't construct with anything, makesure that you call "init"
	HarmMaskProcess() {}

	HarmMaskProcess(T& process, float/*0..1*/fx_val, float/*bin pos*/harm0)
		{ init(process, fx_val, harm0, /*multiply*/1.0f); }

	// override "prepare" to call down with our type (used by the sweepers)
	void prepare(ProcessBase* parent) { _process->prepare(this); }

	// round a frequency down to harmonic centre
	void roundHarm(float bin) {
		_curr_harm = floorf((bin-_offs)/_spacing);
		_curr_cent = _offs+_spacing*_curr_harm;
	}

	// run the process
	void run(long b0, long b1)
	{
		if(b0 < _min_bin) b0 = _min_bin;
		if(b1 > _max_bin) b1 = _max_bin;
		if(b0 > b1) return;

		bool looping = true;
		if(reverse()) {
			// process spectrum in reverse
			roundHarm(b1+_spacing);

			// max bin for harmonic
			long p = b1;
			while(looping) {
				long v1 = (long)(_curr_cent+0.5f)+_half_width;
				long v0 = v1-_width;
				// don't go outside our range
				if(v0 <= b0) {
					v0 = b0;
					looping = false;
				}
				if(v1 > p) v1 = p;
				// process if range is good
				if(v1 >= v0) {
					_process->run(v0, v1);
					p = v0-1;
				}
				_curr_cent -= _spacing;
				_curr_harm--;
			}
		}
		else {
			// process spectrum fwd
			roundHarm((float)b0);

			// min bin for harmonic
			long p = b0;
			while(looping) {
				long v0 = (long)(_curr_cent+0.5f)-_half_width;
				long v1 = v0+_width;
				// don't go outside range
				if(v1 >= b1) {
					v1 = b1;
					looping = false;
				}
				if(v0 < p) v0 = p;
				// process if range is good
				if(v1 >= v0) {
					_process->run(v0, v1);
					p = v1+1;
				}
				_curr_cent += _spacing;
				_curr_harm++;
			}
		}
	}
};

//-------------------------------------------------------------------------------------------------
struct PeakFind
// find peak bin in fft data "x1" between b0 & b1 inclusive
{
	float* max_p;
	float max_pwr;
	long max_bin;

	PeakFind() {}

	PeakFind(pcplxf x1, long b0, long b1) { operator() (x1, b0, b1); }

	void operator () (pcplxf x1, long b0, long b1)
	{
		max_p = x1.r+b0; // arbitrary value if no peak found
		max_pwr = 0.0f;
		for(pcplxf_range x(x1, b0, b1+1); !x.empty(); x++) {
			float t = norm(*x);
			if(t > max_pwr) {
				max_pwr = t;
				max_p = x.r;
			}
		}
		max_bin = max_p-x1.r;
	}

	// by default return peak bin position (or -1 if not found)
	operator long()
	{
		// chk for no peak found
		if(max_pwr == 0.0f) return -1;
		return max_bin;
	}
};

//-------------------------------------------------------------------------------------------------
template <class T> class AutoHarmMaskProcess : public HarmMaskProcess<T>
{
public:
	AutoHarmMaskProcess(T& process, DtBlkFx* b, DtBlkFx::FxParams& params, float fx_val, float multiply)
	{
		long b0 = params.bin[0];
		long b1 = params.bin[1];
		if(b0 > b1) swap(b0, b1);
		long peak_bin = PeakFind(b->FFTdata(), b0, b1);
		/*HarmMaskProcess::*/init(process, fx_val, (float)peak_bin, multiply);
	}
};

//-------------------------------------------------------------------------------------------------
template <class T> class SplitMaskProcess : public MaskProcessBase<T>
{
public:
	// bins from the params
	long _param_bin[2];

	SplitMaskProcess(T& process, DtBlkFx::FxParams& params)
	{
		/*MaskProcessBase::*/_process = &process;
		_param_bin[0] = params.bin[0];
		_param_bin[1] = params.bin[1];
	}

	void run(long b0, long b1)
	{
		if(_param_bin[0] <= _param_bin[1]) {
			// freqA < freqB : normal process inside region
			long v0 = max(b0, _param_bin[0]);
			long v1 = min(b1, _param_bin[1]);
			if(v1 >= v0) _process->run(v0, v1);
		}
		else {
			// freqA > freqB : process outside region
			long v1 = min(b1, _param_bin[1]);
			long v0 = max(b0, _param_bin[0]);
			if(reverse()) {
				if(v1 >= b0) _process->run(b0, v1);
				if(b1 >= v0) _process->run(v0, b1);
			}
			else {
				if(b1 >= v0) _process->run(v0, b1);
				if(v1 >= b0) _process->run(b0, v1);
			}
		}
	}
};

//-------------------------------------------------------------------------------------------------
template <class T> class InvMaskProcess_ : public MaskProcessBase<T>
//
// invert ranges processed
//
{
public:
	// inverted range is limited to this range
	long _min_bin, _max_bin;

	long _prev_bin;

	void init(long min_bin, long max_bin)
	{
		_min_bin = min_bin;
		_max_bin = max_bin;
		_prev_bin = reverse()?_max_bin:_min_bin;
	}

	// do the processing
	void run(long b0, long b1)
	{
		if(reverse()) {
			long v0 = max(b1+1, _min_bin);
			if(v0 <= prev_bin) {
				_process->run(v0, _prev_bin);
				_prev_bin = b0-1;
			}
		}
		else {
			long v1 = min(b0-1, _max_bin);
			if(prev_bin <= v1) {
				_process->run(_prev_bin, v1);
				_prev_bin = b1+1;
			}
		}
	}
	void runFinal()
	{
		if(reverse()) {
			if(_min_bin <= _prev_bin) _process->run(_min_bin, _prev_bin);
		}
		else {
			if(_prev_bin <= _max_bin) _process->run(_prev_bin, _max_bin);
		}
	}
};

//-------------------------------------------------------------------------------------------------
template <class T, class M> class InvMaskProcess : public MaskProcessBase<M>
//
// Processing chain is:
//     InvMaskProcess<T, M<InvMaskProcess_<T> > to M< InvMaskProcess_<T> > to InvMaskProcess_ <T> to T
//
{
	InvMaskProcess_<T> _inv_mask;

public:
	InvMaskProcess(T& process, M& mask)
	{
		_inv_mask._process = process;
		/*MaskProcessBase::*/_process = mask;
	}

	void run(long b0, long b1)
	{
		_inv_mask.init(b0, b1);
		_process->run(b0, b1);
		_inv_mask.runFinal();
	}
};

//-------------------------------------------------------------------------------------------------
template <class T> class ThreshMaskProcess : public MaskProcessBase<T>
//
// ranges correspond to bins above a certain threshold
{
public:
	// fft data from blkfx
	pcplxf _x1;

	float _thresh;

	// width either side of a threshold hit to include
	long _width_bins;

	// width_bins*2 + 1
	long _width2_bins;

	// true=select areas below threshold, false=select above threshold
	bool _select_below;

	ThreshMaskProcess(
		T& process,
		DtBlkFx* b,					//
		float thresh_param,			// param to set width or threshold
		float width_frac			// 
	)
	{
		/*MaskProcessBase::*/_process = &process;
		_x1 = b->FFTdata();

		//_width_bins = (long)(width_frac*(float)b->_freq_fft_n);
		_width_bins = 1;
		_width2_bins = _width_bins*2+1;

		SplitParam<2> split(thresh_param);
		_select_below = split.i_part?true:false;

		// arbitrary threshold function, -40 .. +40 dB above average bin power
		_thresh = powf(10.0f, (split.f_part-0.5f)*4.0f)*b->_total_out_pwr/(float)b->_freq_fft_n;
	}

	bool/*found*/testThresh(cplxf x, bool select_below)
	// select_below=true: return true if "x" is less than the theshold
	// select_below=false: return true if "x" is above the threshold
	{
		if(select_below) {
			if(norm(x) < _thresh) return true;
		}
		else {
			if(norm(x) > _thresh) return true;
		}
		return false;
	}

	bool/*peak found*/findThreshBreakFwd(pcplxf_range& /*in-out*/ r, bool select_below)
	{
		while(!r.empty()) {
			if(testThresh(*r, select_below)) return true;
			r++;
		}
		return false;
	}

	bool/*peak found*/findThreshBreakRev(pcplxf_range& /*in-out*/ r, bool select_below)
	{
		while(!r.empty()) {
			if(testThresh(*r, select_below)) return true;
			r--;
		}
		return false;
	}

	void run(long b0, long b1, bool select_below)
	{
		long v0 = -1, v1 = -1;
		// find points breaking the threshold and include "width" data either side
		if(reverse()) {
			// find first bin to break threshold
			pcplxf_range pk(_x1, b1, b0-1);
			if(findThreshBreakRev(pk, select_below)) v1 = min(pk.r-_x1.r+_width_bins, b1);

			// loop finding points above threshold
			while(!pk.empty()) {
				pk--;
				pcplxf_range next_pk = pk;
				if(!findThreshBreakRev(next_pk, select_below)) break;

				// distance from previous theshold breaker
				long dist = pk.r - next_pk.r;
				if(dist > _width2_bins) {
					// run prev range
					v0 = max(pk.r-_x1.r-_width_bins, b0);
					if(v0 <= v1) _process->run(v0, v1);

					// set start position for this rng
					v1 = next_pk.r-_x1.r+_width_bins;
				}
				pk = next_pk;
			}

			// run final rng
			if(v1 >= 0) {
				v0 = max(pk.r-_x1.r-_width_bins, b0);
				if(v0 <= v1) _process->run(v0, v1);
			}
		}
		else {
			int breaks_found = 0, runs_called = 0;
			// find initial threshold break
			pcplxf_range pk(_x1, b0, b1+1);
			if(findThreshBreakFwd(pk, select_below)) v0 = max(pk.r-_x1.r-_width_bins, b0);

			// loop finding points above threshold
			while(!pk.empty()) {
				pk++;
				pcplxf_range next_pk = pk;
				if(!findThreshBreakFwd(next_pk, select_below)) break;
				breaks_found++;

				// distance from previous theshold breaker
				long dist = next_pk.r - pk.r;
				if(dist > _width2_bins) {
					// run prev range
					v1 = min(pk.r-_x1.r+_width_bins, b1);
					if(v0 <= v1) {
						_process->run(v0, v1);
						runs_called++;
					}

					// set start position for this rng
					v0 = next_pk.r-_x1.r-_width_bins;
				}
				pk = next_pk;
			}

			// run final rng
			if(v0 >= 0) {
				v1 = min(pk.r-_x1.r+_width_bins, b1);
				if(v0 <= v1) {
					_process->run(v0, v1);
					runs_called++;
				}
			}
			DBG("thresh mask" << VAR(breaks_found) << VAR(runs_called));
		}
	}

	void run(long b0, long b1)
	{
		// force inlining by passing "select below" as constant
		if(_select_below) run(b0, b1, /*select below=*/true);
		else run(b0, b1, /*select below=*/false);
	}
};

//*************************************************************************************************
int HarmDispVal(char* /*out*/text, float/*0..1*/val)
{
	const char* name[] = {"both", "odd", "even", "between"};
	SplitParam<4> h(val);
	return sprintf(text, "%.0f%% %s", h.f_part*100.0f, name[h.i_part]);
}

//-------------------------------------------------------------------------------------------------
class HarmMaskFx : public FFTFx
// harmonic mask effect does no processing - it provides a freq mask for the next effect
{
public:
	HarmMaskFx(const char* name, bool freq_b_param_used) : FFTFx(name)
	{
		_params_used[BlkFxParam::FX_FREQ_B] = freq_b_param_used;
		_params_used[BlkFxParam::FX_AMP] = false;
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return HarmDispVal(text, val); }
};
HarmMaskFx g_harm_mask("HarmMask", /*freq_b_param_used*/false);
HarmMaskFx g_auto_harm_mask("AutoHarmMask", /*freq_b_param_used*/true);
HarmMaskFx g_auto_subharm1_mask("ASubH1Mask", /*freq_b_param_used*/true);
HarmMaskFx g_auto_subharm2_mask("ASubH2Mask", /*freq_b_param_used*/true);
HarmMaskFx g_auto_subharm3_mask("ASubH3Mask", /*freq_b_param_used*/true);

//*************************************************************************************************
int ThreshDispVal(char* /*out*/text, float/*0..1*/val)
{
	const char* type[2] = { "above", "below" };
	SplitParam<2> split(val);
	return sprintf(text, "%s %2.0f%%", type[split.i_part], split.f_part*100.0f);
}

//-------------------------------------------------------------------------------------------------
class ThreshMaskFx : public FFTFx
{
public:
	ThreshMaskFx() : FFTFx("ThreshMask")
	{
		_params_used[BlkFxParam::FX_AMP] = false;
		_params_used[BlkFxParam::FX_FREQ_A] = false;
		_params_used[BlkFxParam::FX_FREQ_B] = false;
	}

	virtual void process(DtBlkFx* b) {}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return ThreshDispVal(text, val); }

} g_thresh_mask;


//-------------------------------------------------------------------------------------------------
template<class T> inline void run_(DtBlkFx* b, T& process, long b0, long b1)
// run "process" over bins b0->b1 & adjust pwr
{
	process.prepare(/*parent*/NULL);
	if(b0 <= b1) process.run(b0, b1);
	process.done();
	b->_total_out_pwr += process.pwrChg();
}

//-------------------------------------------------------------------------------------------------
template<class T> inline void run_(DtBlkFx* b, T& process, bool use_split_mask, long b0, long b1)
{
	if(use_split_mask) run_(b, SplitMaskProcess<T>(process, b->currFxParams()), b0, b1);
	else run_(b, process, b0, b1);
}

//-------------------------------------------------------------------------------------------------
template<class T> inline bool/*ran mask*/runHarmMask(
	DtBlkFx* b,
	T& end_process,
	bool use_split_mask,
	long b0,
	long b1
)
{
	// check what the prev params were
	int fx_set = b->_fx_set;
	if(fx_set <= 0) return false;

	DtBlkFx::FxParams& prev_params = b->_fx_params[fx_set-1];
	FFTFx *prev_fx = prev_params.fft_fx;

	if(prev_fx == &g_harm_mask) {
		run_(b,
			HarmMaskProcess<T>(end_process, prev_params.val, prev_params.fbin[0]),
			use_split_mask,
			b0, b1
		);
		return true;
	}

	float mult;
	if(prev_fx == &g_auto_harm_mask) mult = 1.0f;
	else if(prev_fx == &g_auto_subharm1_mask) mult = 0.5f;
	else if(prev_fx == &g_auto_subharm2_mask) mult = 1.0f/3.0f;
	else if(prev_fx == &g_auto_subharm3_mask) mult = 0.25f;
	else return false;

	AutoHarmMaskProcess<T> autoharm(end_process, b, prev_params, prev_params.val, mult);
	run_(b, autoharm, use_split_mask, b0, b1);
	return true;
}

//-------------------------------------------------------------------------------------------------
template<class T> inline void run(
	DtBlkFx* b,
	T& end_process,
	bool use_split_mask,
	long b0,
	long b1
)
//
// default BlkFx process calls 
//
{
	// check what the prev params were
	int fx_set = b->_fx_set;

	if(fx_set > 0) {
		DtBlkFx::FxParams& prev_params = b->_fx_params[fx_set-1];
		if(runHarmMask(b, end_process, use_split_mask, b0, b1)) return;

		// check for other masks
		if(prev_params.fft_fx == &g_thresh_mask) {
			run_(b,
				ThreshMaskProcess<T>(end_process, b, prev_params.val, prev_params.freq_param[0]*.05f),
				use_split_mask,
				b0, b1
			);
			return;
		}
	}

	// no previous mask effect, do normal process
	run_(b, end_process, use_split_mask, b0, b1);
}

// run over entire range
template<class T> inline void run(DtBlkFx* b, T& end_process, bool use_split_mask=true)
	{ run(b, end_process, use_split_mask, /*start bin*/1, /*end bin*/b->_freq_fft_n/2-1); }


//*************************************************************************************************
class HarmFiltFx : public FFTFx
{
public:
	HarmFiltFx() : FFTFx("HarmFilt") {}

	virtual void process(DtBlkFx* b)
	{
		DtBlkFx::FxParams& params = b->currFxParams();
		AmpProcess amp(b);
		HarmMaskProcess<AmpProcess> harm(
			amp,
			params.val, // value
			min(params.fbin[0], params.fbin[1]) // centre
		);
		run(b, harm, /*split mask*/false, /*start bin*/1, /*end bin*/max(params.bin[0], params.bin[1]));
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return HarmDispVal(text, val); }

} g_harm_filt_fx;



//*************************************************************************************************
// no effect does nothing
class NoFx : public FFTFx
{
public:
	NoFx() : FFTFx("NoEffect")
	{
		memset(_params_used, 0, sizeof(_params_used));
	}
} g_no_fx;

//*************************************************************************************************
// filter effect
class FilterFx : public FFTFx
{
public:
	FilterFx() : FFTFx("Filter")
	{
		_params_used[BlkFxParam::FX_VAL] = false;
	}
	virtual void process(DtBlkFx* b) { run(b, AmpProcess(b)); }

} g_filter_fx;


//*************************************************************************************************
struct ContrastProcess : public AmpProcess
// "contrast" effect performs abs(x) ^ (2*raise+1) over the spectrum
//
{
	static float scaleVal(float v)
	{
		v -= 0.5f;
		if(v >= 0.0f) return lin_interp(v*v*4.0f, 0.0f, 4.0f);
		else return lin_interp(v*v*4.0f, 0.0f, -0.5f);
	}


	// pwr to raise to
	float _raise;

	float _scale;

	ContrastProcess(DtBlkFx* b) : AmpProcess(b)
	{
		_raise = scaleVal(b->currFxParams().val);

		// find a scale value to make the total spectrum pwr 10.0f (arbitrary)
		if(b->_total_out_pwr > 1e-20) _scale = 10.0f/b->_total_out_pwr;
		else _scale = 1.0f;
	}

	void run(long b0, long b1)
	{
		// input power for this range
		float in_pwr = 0;

		// output power for this range
		float out_pwr = 0;

		// variables from AmpProcess: _x1, _in_pwr, _amp
		if(_raise >= 0) {
			// raise positive
			for(pcplxf_range x(_x1, b0, b1+1); !x.empty(); x++) {
				float t = norm(*x); 
				in_pwr += t;
				*x = (*x) * powf(t*_scale, _raise); // equivalent to abs(*x) = abs(*x)^(2*raise+1) 
				out_pwr += norm(*x);
			}
		}
		else {
			// raise negative
			for(pcplxf_range x(_x1, b0, b1+1); !x.empty(); x++) {
				float t = norm(*x); 
				in_pwr += t;
				if(t > 1e-20)
					*x = (*x) * powf(t*_scale, _raise); // equivalent to abs(*x) = abs(*x)^(2*raise+1) 
				out_pwr += norm(*x);
			}
		}

		// match output pwr to input power
		MatchPwr(_amp, in_pwr, out_pwr, pcplxf_range(_x1, b0, b1+1));

		// 
		_in_pwr += in_pwr;
	}

};

//-------------------------------------------------------------------------------------------------
class ContrastFx : public FFTFx
{
public:
	ContrastFx() : FFTFx("Contrast") { }

	virtual void process(DtBlkFx* b) { run(b, ContrastProcess(b)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return sprintf(text, "%.0f", (val-0.5f)*200.0f); }

} g_contrast_fx;


//******************************************************************************************
struct SmearProcess : public AmpProcess
{
	float _smear;

	SmearProcess(DtBlkFx* b) : AmpProcess(b)
	{
		_smear = b->currFxParams().val;
	}

	void run(long b0, long b1)
	//
	// randomize the phase
	//
	{
		long rand_i = g_rand_i;
		for(pcplxf_range x(/*AmpProcess::*/_x1, b0, b1+1); !x.empty(); x++) {
			/*AmpProcess::*/_in_pwr += norm(*x);
			*x = /*AmpProcess::*/_amp * (*x) * (sincos_table[rand_i]*_smear+1.0f-_smear);
			rand_i = prbs29(rand_i);
		}

		g_rand_i = rand_i;
	}

};


//-------------------------------------------------------------------------------------------------
class SmearFx : public FFTFx
{
public:
	SmearFx() : FFTFx("Smear") {}

	virtual void process(DtBlkFx* b) { run(b, SmearProcess(b)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return sprintf(text, "%2.0f%%", val*100.0f); }
} g_smear_fx;



//-------------------------------------------------------------------------------------------------
class ThreshFx : public FFTFx
{
public:
	ThreshFx() : FFTFx("Thresh") {}

	virtual void process(DtBlkFx* b)
	{
		DtBlkFx::FxParams& params = b->currFxParams();
		AmpProcess amp(b);
		ThreshMaskProcess<AmpProcess> thresh(amp, b, /*thresh param*/params.val, /*width frac*/0.003f);
		run(b, thresh);
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return ThreshDispVal(text, val); }

} g_thresh_fx;


//*************************************************************************************************
struct ClipProcess : public AmpProcess
{

	// threshold for clipping
	float _thresh;

	// sqrt thresh
	float _root_thresh;

	ClipProcess(DtBlkFx* b) : AmpProcess(b)
	{
		// arbitary threshold function
		_thresh = 1.0f - b->currFxParams().val;
		_thresh *= _thresh*8.0f*b->_total_out_pwr/b->_freq_fft_n;
		if(_thresh < 1e-37f) _thresh = 1e-37f;

		_root_thresh = sqrtf(_thresh);
	}

	void run(long b0, long b1)
	// clip bins with magnitude higher than the threshold
	{
		float in_pwr = 0;
		float out_pwr = 0;
		for(pcplxf_range x(/*AmpProcess::*/_x1, b0, b1+1); !x.empty(); x++) {
			float t0 = norm(*x);
			in_pwr += t0;
			if(t0 >= _thresh) {
				float t1 = _root_thresh / sqrtf(t0);
				*x = (*x) * t1;
				out_pwr += _thresh;
			}
			else out_pwr += t0;
		}
		MatchPwr(_amp, in_pwr, out_pwr, pcplxf_range(_x1, b0, b1+1));
		/*AmpProcess::*/_in_pwr += in_pwr;
	}
};

//-------------------------------------------------------------------------------------------------
class ClipFx : public FFTFx
{
public:
	ClipFx() : FFTFx("Clip") {}

	virtual void process(DtBlkFx* b) { run(b, ClipProcess(b)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return sprintf(text, "%2.0f%%", val*100.0f); }

} g_clip_fx;


//-------------------------------------------------------------------------------------------------
static float/*octaves shift*/GetOctShift(float/*0..1*/v)
	{ return lin_interp(v, -3.0f, 3.0f); }

//*************************************************************************************************
static float/*Hz*/GetFreqShift(float v/*0..1*/)
//
// Convert freq shift param to Hz
//
// return frequency as a fraction between -22050 to 22050 Hz
// maps the linear input "v" to an arbitary linear+exponential function
// (linear transitioning into exponential)
//
{
	const float lin = 2048.0f/32768.0f;
	const int exp_raise = 8;
	const float exp_scale = (1.0f-lin)/(1<<exp_raise);
	v = (v-0.5f)*2;

	bool pos = true;
	if(v < 0) {
		v = -v;
		pos = false;
	}
	float r = (powf(2, v*exp_raise)-1.0f)*exp_scale + v*lin;
	return (pos ? r:-r)*22050.0f;
}
//-------------------------------------------------------------------------------------------------
template <class T/*dummy class type*/> class ShiftProcess : public AmpProcess
// "contrast" effect performs abs(x) ^ (2*raise+1) over the spectrum
//
{
public:
	// power change
	float _pwr_chg;

	// max bin
	long _max_bin;

	// true=constant shift, false=scale shift
	bool _do_const_shift;

	// bin range meaning: true=destination, false=source
	bool _dest_mode;

	// bin shift distance
	long _const_bin_shift;

	// constant shift phase correction and amp multiply
	cplxf _const_phase_amp;

	// variables only used for scaled freq shift (_do_const_shift=false)

	// inverse scale mult used for dest_mode only
	float _inv_freq_scale;

	// centre freq multiplier to do scaled shift
	float _scale_mult;

	// fft blk len
	long _fft_len;

	float _bin_to_hz, _hz_to_bin;

	// lowest bin that we will do type a shifting
	long _type_a_low_bin;
	long _type_b_width;
	

	// direction to run processing
	bool _reverse;
	bool reverse() { return _reverse; }

	// absolute sample position multiplied by sincos table length
	long long _scaled_samp_abs;

	cplxf getPhaseAmp(double bin_shift)
	// find phase correction for the 
	{
		return sincos_table[
			(int)(_scaled_samp_abs*bin_shift/_fft_len)
		] * AmpProcess::_amp;
	}

	//
	ShiftProcess(DtBlkFx* b, bool dest_mode) : AmpProcess(b)
	{
		_pwr_chg = 0;
		_dest_mode = dest_mode;

		_fft_len = b->_freq_fft_n;
		_max_bin = _fft_len/2-1;
		_scaled_samp_abs = b->_blk_samp_abs*sincos_table.LEN;

		// split param into 2
		SplitParam<2> split(b->currFxParams().val);

		// process in reverse when shifting up in frequency

		if(split.i_part) {
			// constant frequency shift
			_do_const_shift = true;
			_const_bin_shift = (long)(
				GetFreqShift(split.f_part)/b->getSampleRate()*(float)_fft_len + 0.5f
			);
			_const_phase_amp = getPhaseAmp(_const_bin_shift);
			_reverse = (_const_bin_shift > 0);
		}
		else {
			// scaled frequency shift
			_do_const_shift = false;
			float freq_scale = powf(2.0f, GetOctShift(split.f_part));
			_scale_mult = freq_scale-1.0f;
			_reverse = (freq_scale > 1.0f);
			if(_dest_mode) _inv_freq_scale = 1.0f/freq_scale;

			_bin_to_hz = _b->getSampleRate()/(float)_fft_len;
			_hz_to_bin = (float)_fft_len/_b->getSampleRate();

			// arbitrary sizes
			_type_a_low_bin = _fft_len/130;
			_type_b_width = max(_fft_len/700, 1L);
		}
	}

	// run method is dependent on "T"
	void run(pcplxf dst, pcplxf src, cplxf phase_amp);

	void run(long dst_bin, long src_bin, long n_bins, cplxf phase_amp)
	{
		if(src_bin < 1) {
			dst_bin += (1-src_bin);
			n_bins -= (1-src_bin);
			src_bin = 1;
		}
		if(src_bin+n_bins > _max_bin) n_bins = _max_bin - src_bin;

		if(dst_bin < 1)  {
			src_bin += (1-dst_bin);
			n_bins -= (1-dst_bin);
			dst_bin = 1;
		}
		if(dst_bin+n_bins > _max_bin) n_bins = _max_bin - dst_bin;

		if(n_bins < 1) return;

		if(_reverse) {
			// process spectrum in reverse (shift up)
			pcplxf xd(_x1, dst_bin+n_bins);
			for(pcplxf_range xs(_x1, src_bin+n_bins-1, src_bin-1); !xs.empty(); xs--, xd--) {
				_pwr_chg -= norm(*xd);
				run(xd, xs, phase_amp);
				_pwr_chg += norm(*xd);
			}
		}
		else {
			// process spectrum forwards (shift down)
			pcplxf xd(_x1, dst_bin);
			for(pcplxf_range xs(_x1, src_bin, src_bin+n_bins); !xs.empty(); xs++, xd++) {
				_pwr_chg -= norm(*xd);
				run(xd, xs, phase_amp);
				_pwr_chg += norm(*xd);
			}
		}
	}
	void shiftUpA(long b0, long b1)
	{
		float hz = (float)b1*_bin_to_hz;
		float part_note = 48.0f*NoteFreq::inv_log_c0_hz*(logf(hz)-NoteFreq::log_c0_hz);
		float cf = NoteFreq::c0_hz*powf(2.0f, floorf(part_note)*(1.0f/48.0f));
		float cb = cf*_hz_to_bin;

		long v1 = b1;
		while(1) {
			float fshift = floorf(cb*_scale_mult+0.5f);
			float next_cb = cb*(float)0.98566319864018758778;
			long v0 = (long)((next_cb+cb)*0.5f);
			if(v0 <= v1) {
				run(v0+(long)fshift, v0, v1-v0+1, getPhaseAmp(fshift));
				v1 = v0-1;
				if(v1 < b0) break;
			}
			cb = next_cb;
		}
	}

	void shiftDownA(long b0, long b1)
	{
		float hz = (float)b0*_bin_to_hz;
		float part_note = 48.0f*NoteFreq::inv_log_c0_hz*(logf(hz)-NoteFreq::log_c0_hz);
		float cf = NoteFreq::c0_hz*powf(2.0f, ceilf(part_note)*(1.0f/48.0f));
		float cb = cf*_hz_to_bin;

		long v0 = b0;
		while(1) {
			float fshift = floorf(cb*_scale_mult+0.5f);
			float next_cb = cb*(float)1.0145453349375237462;
			long v1 = (long)((next_cb+cb)*0.5f);
			if(v0 <= v1) {
				run(v0+(long)fshift, v0, v1-v0+1, getPhaseAmp(fshift));
				v0 = v1+1;
				if(v0 > b1) break;
			}
			cb = next_cb;
		}
	}

	void shiftUpB(long b0, long b1)
	{
		// try to align shifts with peaks in the data
		long v0[3], v1[3];
		PeakFind p[3];
		v1[0] = b1;
		while(1) {
			v0[0] = max(v1[0]-_type_b_width, b0);
			v1[1] = v0[0]-1;
			v0[1] = max(v1[1]-_type_b_width, b0);
			v1[2] = v0[1]-1;
			v0[2] = max(v1[2]-_type_b_width, b0);

			for(int i = 0; i < 3; i++) p[i](_x1, v0[i], v1[i]);

			// decide whether to use p[0] or p[1] as the peak
			long cent;
			if(p[1].max_pwr > p[0].max_pwr && p[1].max_pwr > p[2].max_pwr)
				cent = p[1].max_bin;
			else cent = p[0].max_bin;

			float fbin_shift = floorf((float)cent*_scale_mult+0.5f);
			v0[0] = max(cent-_type_b_width, b0);
			run(v0[0]+(long)fbin_shift, v0[0], v1[0]-v0[0]+1, getPhaseAmp(fbin_shift));

			if(v0[0] <= b0) break;
			v1[0] = v0[0]-1;
		}
	}

	void shiftDownB(long b0, long b1)
	{
		// try to align shifts with peaks in the data
		long v0[3], v1[3];
		PeakFind p[3];
		v0[0] = b0;
		while(1) {
			v1[0] = min(v0[0]+_type_b_width, b1);
			v0[1] = v1[0]+1;
			v1[1] = min(v0[1]+_type_b_width, b1);
			v0[2] = v1[1]+1;
			v1[2] = min(v0[2]+_type_b_width, b1);

			for(int i = 0; i < 3; i++) p[i](_x1, v0[i], v1[i]);

			// decide whether to use p[0] or p[1] as the peak
			long cent;
			if(p[1].max_pwr > p[0].max_pwr && p[1].max_pwr > p[2].max_pwr)
				cent = p[1].max_bin;
			else cent = p[0].max_bin;

			float fbin_shift = floorf((float)cent*_scale_mult+0.5f);
			v1[0] = min(cent+_type_b_width, b1);
			run(v0[0]+(long)fbin_shift, v0[0], v1[0]-v0[0]+1, getPhaseAmp(fbin_shift));

			if(v1[0] >= b1) break;
			v0[0] = v1[0]+1;
		}
	}

	void run(long b0, long b1)
	{

		if(_do_const_shift) {
			// constant shift is easy
			long n_bins = b1-b0+1;
			long src_bin, dst_bin;
			
			if(_dest_mode) {
				dst_bin = b0;
				src_bin = b0-_const_bin_shift;
			}
			else {
				src_bin = b0;
				dst_bin = b0+_const_bin_shift;
			}
			run(dst_bin, src_bin, n_bins, _const_phase_amp);
		}
		else {
			// scaled shift is messy

			if(_dest_mode) {
				b1 = (long)((float)b1*_inv_freq_scale+0.5f);
				b0 = (long)((float)b0*_inv_freq_scale+0.5f);
			}

			if(b1 < 1) return;
			if(b0 < 1) b0 = 1;

			if(_reverse) {
				// shift up
				if(b1 >= _type_a_low_bin)
					shiftUpA(max(_type_a_low_bin, b0), b1);
				if(b0 < _type_a_low_bin)
					shiftUpB(b0, min(_type_a_low_bin-1, b1));
			}
			else {
				// shift down
				if(b0 < _type_a_low_bin)
					shiftDownB(b0, min(_type_a_low_bin-1, b1));
				if(b1 >= _type_a_low_bin)
					shiftDownA(max(_type_a_low_bin, b0), b1);
			}
		}
	}

	float pwrChg() { return _pwr_chg; }
};


//-------------------------------------------------------------------------------------------------
int ShiftDispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val)
{
	SplitParam<2> split(val);
	if(split.i_part) {
		// constant freq shift
		float hz = GetFreqShift(split.f_part);
		return sprintf(
			text,
			"%.0f Hz", 
			b? b->guessRoundHz(hz):hz
		);
	}
	else {
		// scaled shift
		return sprintf(text, "%.2f notes",  GetOctShift(split.f_part)*12.0f);
	}
}


//- SHIFT ADD -------------------------------------------------------------------------------------
class ShiftAddFx;
template <> inline void ShiftProcess<ShiftAddFx>::run(pcplxf dst, pcplxf src, cplxf phase_amp)
	{ *dst = *src * phase_amp + *dst; }

//-------------------------------------------------------------------------------------------------
class ShiftAddFx : public FFTFx
{
public:
	bool _dest_mode;

	ShiftAddFx(const char* name, bool dest_mode) : FFTFx(name)
		{ _dest_mode = dest_mode; }

	virtual void process(DtBlkFx* b) { run(b, ShiftProcess<ShiftAddFx>(b, _dest_mode)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return ShiftDispVal(b, text, val); }
};

ShiftAddFx g_shift_add_fx("ShiftAdd", /*dst mode*/false);
ShiftAddFx g_shift_addd_fx("ShiftAddD", /*dst mode*/true);


//- SHIFT REPL ------------------------------------------------------------------------------------
class ShiftReplFx;
template <> inline void ShiftProcess<ShiftReplFx>::run(pcplxf dst, pcplxf src, cplxf phase_amp)
	{ *dst = *src * phase_amp; }

//-------------------------------------------------------------------------------------------------
class ShiftReplFx : public FFTFx
{
public:
	bool _dest_mode;

	ShiftReplFx(const char* name, bool dest_mode) : FFTFx(name)
		{ _dest_mode = dest_mode; }

	virtual void process(DtBlkFx* b) { run(b, ShiftProcess<ShiftReplFx>(b, _dest_mode)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return ShiftDispVal(b, text, val); }
};

ShiftReplFx g_shift_repl_fx("ShiftRepl", /*dst mode*/false);
ShiftReplFx g_shift_repld_fx("ShiftReplD", /*dst mode*/true);

//- SHIFT MOVE ------------------------------------------------------------------------------------
class ShiftMoveFx;
template <> inline void ShiftProcess<ShiftMoveFx>::run(pcplxf dst, pcplxf src, cplxf phase_amp)
{
	_pwr_chg -= norm(*src);
	cplxf s = *src;
	*src = cplxf(0, 0);
	*dst = *dst + s * phase_amp;
}

//-------------------------------------------------------------------------------------------------
class ShiftMoveFx : public FFTFx
{
public:
	bool _dest_mode;

	ShiftMoveFx(const char* name, bool dest_mode) : FFTFx(name)
		{ _dest_mode = dest_mode; }

	virtual void process(DtBlkFx* b) { run(b, ShiftProcess<ShiftMoveFx>(b, _dest_mode)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return ShiftDispVal(b, text, val); }
};

ShiftMoveFx g_shift_move_fx("ShiftMove", /*dst mode*/false);
ShiftMoveFx g_shift_moved_fx("ShiftMoveD", /*dst mode*/true);

//*************************************************************************************************
class AutoHarmFx : public FFTFx
{
public:
	AutoHarmFx() : FFTFx("AutoHarm") {}

	virtual void process(DtBlkFx* b)
	{
		DtBlkFx::FxParams& params = b->currFxParams();
		long stop_bin = b->_freq_fft_n/2-1;
		if(params.bin[0] > params.bin[1]) stop_bin = params.bin[0];

		AmpProcess amp(b);
		AutoHarmMaskProcess<AmpProcess> harm(amp, b, params, params.val, /*multiplier*/1.0f);
		run(b, harm, /*split mask*/false, /*start_bin*/1, stop_bin);
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return HarmDispVal(text, val); }

} g_auto_harm_fx;


//*************************************************************************************************
class MirrorProcess : public AmpProcess
{
public:
	float _pwr_chg;
	float _s[2];

	MirrorProcess(DtBlkFx* b) : AmpProcess(b)
	{
		_pwr_chg = 0;
		float v = b->currFxParams().val;
		_s[0] = _amp * (1-v);
		_s[1] = _amp * v;
	}

	void run(long b0, long b1)
	{
		pcplxf xa = _x1+b0;
		pcplxf xb = _x1+b1;
		while(xa.r <= xb.r) {
			_pwr_chg -= norm(*xa);
			_pwr_chg -= norm(*xb);
			cplxf ta = *xa;
			cplxf tb = *xb;
			*xa = ta*_s[0] + tb*_s[1];
			*xb = tb*_s[0] + ta*_s[1];
			_pwr_chg += norm(*xa);
			_pwr_chg += norm(*xb);
			xa++;
			xb--;
		}
	}

	float pwrChg() { return _pwr_chg; }
};


//-------------------------------------------------------------------------------------------------
class MirrorFx : public FFTFx
{
public:
	MirrorFx() : FFTFx("Mirror") { }
	virtual void process(DtBlkFx* b) { run(b, MirrorProcess(b)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
		{ return sprintf(text, "%.0f%%", val*100.0f); }

} g_mirror_effect;


//*************************************************************************************************
//-------------------------------------------------------------------------------------------------
class HarmMatchProcess : public AmpProcess
{
public:
	// 2 harmonic sequences that we linearly interpolate against
	float* _harms[2];

	// number of harmonics per sequence
	int _n_harms;

	// scale factors (0..1) to apply to harm sequences
	float _harms_interp[2];

	// amount effect has changed power 
	float _pwr_chg;

	// power scaling to apply to harmonics
	float _pwr_scale;

	// mix original data
	bool _mix_mode;

	// number of bins in fundamental
	int _fund_len;

	// copy of fundamental data if we're in copy mode (NULL in scale mode)
	pcplxf _fund_copy;

	HarmMaskProcess<HarmMatchProcess> *_parent;

	void prepare(HarmMaskProcess<HarmMatchProcess>* parent)
	{
		_parent = parent;

		long max_bin = _b->_freq_fft_n/2-1;

		// determine bin range of fundamental
		long b0 = limit_range(
			(long)(parent->_freq - parent->_half_width), 1, max_bin);

		long b1 = min(b0+(long)parent->_width, max_bin);

		if(b0 > b1) {
			// abnormal case, should not happen
			_fund_len = 0;
			_pwr_scale = 1.0f;
			return;
		}

		_fund_len = b1-b0+1;
		float fund_pwr = 0.0f;
		if(_fund_copy.r) {
			// take a copy of the fundamental scaled by required amp
			_fund_copy.i = _fund_copy.r+_fund_len*2-1;
			pcplxf f = _fund_copy;
			for(pcplxf_range x(_x1, b0, b1+1); !x.empty(); x++) {
				fund_pwr += norm(*x);
				*f = (*x) * _amp;
				f++;
			}
		}
		else {
			// scale mode
			fund_pwr = GetPwr(pcplxf_range(_x1, b0, b1+1));
		}

		float harm0_pwr = harmPwr(0);
		if(harm0_pwr > 0) _pwr_scale = fund_pwr/harm0_pwr;
	}

	// return interpolated harmonic power from table
	float harmPwr(int harmonic)
		{ return _harms[0][harmonic]*_harms_interp[0]+_harms[1][harmonic]*_harms_interp[1]; }


	HarmMatchProcess(
		DtBlkFx* b,
		HarmData harm_data
	) : AmpProcess(b)
	{
		
		_pwr_chg = 0;

		_n_harms = harm_data.n_harms;

		// grab the fx params
		DtBlkFx::FxParams& params = b->currFxParams();
		SplitParam<2> val(params.val);

		// in copy mode we use _b->_x2 for temporary storage of the harmonic that
		// we'll copy
		if(val.i_part) _fund_copy = pcplxf(&(_b->_x2[0]), 0L); // copy mode
		else _fund_copy = pcplxf(NULL, 0L); // scale mode
		
		// find which harmonic sequences to interpolate from
		int max_table = harm_data.n_tables-1;
		float ftable_i;
		_harms_interp[1] = modff(val.f_part*(float)max_table, &ftable_i);
		_harms_interp[0] = 1-_harms_interp[1];

		int table_i = limit_range((int)ftable_i, 0, max_table);
		_harms[0] = harm_data.data+table_i*harm_data.n_harms;

		if(table_i < max_table) table_i++;
		_harms[1] = harm_data.data+table_i*harm_data.n_harms;

		// determine whether we need to mix or replace
		_mix_mode = (_amp < 1.0f);

	}

	void run(long b0, long b1)
	{
		long curr_harm = (long)_parent->_curr_harm;
		if(curr_harm < 0 || curr_harm >= _n_harms) return;

		// get required power for this harmonic
		float harm_pwr = harmPwr(curr_harm);

		if(_fund_copy.r) {
			// copy mode
			pcplxf f = _fund_copy;

			// power change due to this effect
			float pwr_chg = 0;

			// copy scaling due to harmonic
			float scale = sqrtf(harm_pwr);

			// find bin-limits for harmonic
			long s0 = (long)(_parent->_curr_cent+0.5f) - _fund_len/2;
			long s1 = s0 + _fund_len-1;

			// adjust bins and start of copy if need be
			if(b0 < s0) b0 = s0;
			else f = f+(b0-s0);
			if(b1 > s1) b1 = s1;
			if(b0 > b1) return;

			if(_mix_mode) {
				// mix mode, mix source & destination
				// use scaling on source as per "scale mode"
				float orig_pwr = GetPwr(pcplxf_range(_x1, b0, b1+1));
				float target_pwr = _pwr_scale*harm_pwr;
				float orig_scale = MatchPwr(1.0f-_amp, target_pwr, orig_pwr);

				for(pcplxf_range x(_x1, b0, b1+1); !x.empty(); x++) {
					pwr_chg -= norm(*x);
					*x = (*x) * orig_scale + (*f) * scale;
					pwr_chg += norm(*x);
					f++;
				}
			}
			else {
				// not mix mode, replace destination
				for(pcplxf_range x(_x1, b0, b1+1); !x.empty(); x++) {
					pwr_chg -= norm(*x);
					*x = (*f) * scale;
					pwr_chg += norm(*x);
					f++;
				}
			}
			_pwr_chg += pwr_chg;
		}
		else {
			// scale mode

			// find what we need to scale the existing data by to match the harmonic power
			float orig_pwr = GetPwr(pcplxf_range(_x1, b0, b1+1));
			float target_pwr = _pwr_scale*harm_pwr;

			// attempt to match to harmonic pwr
			float scale;
			if(_mix_mode)
				scale = 1+_amp*(MatchPwr(1.0f, target_pwr, orig_pwr)-1);
			else
				scale = MatchPwr(_amp, target_pwr, orig_pwr);

			for(pcplxf_range x(_x1, b0, b1+1); !x.empty(); x++) *x = (*x) * scale;

			// store power change
			_pwr_chg += (scale*scale-1)*orig_pwr;
		}
	}

	float pwrChg() { return _pwr_chg; }
};


//-------------------------------------------------------------------------------------------------
class HarmMatchFx : public FFTFx
{
public:
	HarmData _data;
	HarmMatchFx(const char* name, HarmData data) : FFTFx(name) { _data = data; }

	virtual void process(DtBlkFx* b)
	{
		// if the previous effect is a harmonic mask then use it (so that the user can control
		// harmonic width) otherwise default to processing all harmonics with a width of 50%
		HarmMatchProcess match(b, _data);
		// attemp to use previous effect as harmonic mask
		if(!runHarmMask(b, match,/*split mask*/true, 
				/*start bin*/1, /*stop bin*/b->_freq_fft_n/2-1)) {

			// if the previous effect wasn't a harmonic mask, create one using a constant fx val param
			AutoHarmMaskProcess<HarmMatchProcess>
				auto_match(match, b, b->currFxParams(), /*fx_val*/0.125f, /*freq mult*/1.0f);
			
			// run without slit mask since we are using freq's for doing the auto-harm
			run(b, auto_match, /*split mask*/false);
		}
	}

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
	{
		SplitParam<2> v(val);
		return sprintf(text, "%s/%.0f", v.i_part?"copy":"cabbage", v.f_part*100);
	}

	virtual bool ampMixMode() { return true; }
};

extern HarmData sweep1_coeff;
HarmMatchFx g_sweep1_fx("Triangles", sweep1_coeff);

extern HarmData sweep2_coeff;
HarmMatchFx g_sweep2_fx("Squares", sweep2_coeff);

extern HarmData sweep3_coeff;
HarmMatchFx g_sweep3_fx("Saws", sweep3_coeff);

extern HarmData sweep4_coeff;
HarmMatchFx g_sweep4_fx("Pointy", sweep4_coeff);

extern HarmData sweep5_coeff;
HarmMatchFx g_sweep5_fx("Sweep", sweep5_coeff);


//*************************************************************************************************
class ResampleProcess : public AmpProcess
{
public:
	// max bin that can be processed without overflowing
	long _max_bin;

	// frequency scale
	float _freq_mult;
	FixPoint _freq_mult_fix;

	// output pwr
	float _pwr_chg;

	// true if we are shifting up in frequency
	bool _shift_up;

	// process in reverse
	bool reverse() { return _shift_up; }

	// absolute sample position multiplied by sincos table length
	long long _scaled_samp_abs;

	// fft blk len
	long _fft_len;

	cplxf getPhaseAmp(double bin_shift)
	// find phase correction for the 
	{
		return sincos_table[
			(int)(_scaled_samp_abs*bin_shift/_fft_len)
		] * AmpProcess::_amp;
	}

	ResampleProcess(DtBlkFx* b) : AmpProcess(b)
	{
		_fft_len = b->_freq_fft_n;
		_scaled_samp_abs = b->_blk_samp_abs*sincos_table.LEN;

		// get multiplier from octave shift
		_freq_mult = powf(2.0f, GetOctShift(b->currFxParams().val));
		_freq_mult_fix = _freq_mult;

		// maximum bin that can be processed without exceeding fft length
		_max_bin = (long)((float)(_fft_len/2-1)/_freq_mult);

		_shift_up = (_freq_mult > 1.0f);

		_pwr_chg = 0.0f;

	}

	void run(pcplxf dst, pcplxf src, cplxf phase_amp)
	{
		// remove src
		cplxf s = *src;
		//*src = cplxf(0,0);
		//_pwr_chg -= norm(s);

		// move src to dst
		_pwr_chg -= norm(*dst);

		*dst = (*dst) + s*phase_amp;
		_pwr_chg += norm(*dst);
	}

	void run(long b0, long b1)
	{
		// variables from AmpProcess: _x1
		if(_shift_up) {
			// shifting up in frq, process backwards

			// make sure that we don't overflow
			if(b1 > _max_bin) b1 = _max_bin;

			// see if we've still got something to process
			if(b0 > b1) return;

			float d_fix = (float)b1*_freq_mult;

			for(pcplxf_range x(_x1, b1, b0-1); !x.empty(); x--, b1--) {
				long dst = (long)d_fix;
				run(_x1+dst, x, getPhaseAmp(dst-b1));
				d_fix -= _freq_mult;	
			}
		}
		else {
			// shifting down in frq, process forward
			float d_fix = (float)b0*_freq_mult;
			for(pcplxf_range x(_x1, b0, b1+1); !x.empty(); x++, b0++) {
				long dst = (long)d_fix;
				run(_x1+dst, x, getPhaseAmp(dst-b0));
				d_fix += _freq_mult;
			}
		}
	}

	float pwrChg() { return _pwr_chg; }
};

//-------------------------------------------------------------------------------------------------
class ResampleFx : public FFTFx
{
public:
	ResampleFx() : FFTFx("Resample") {}

	virtual void process(DtBlkFx* b) { run(b, ResampleProcess(b)); }

	virtual int/*num chars written*/dispVal(DtBlkFx* b, char* /*out*/text, float/*0..1*/val, int fx_set)
	{
		// display octave shift as notes (semitones) shift
		float notes = GetOctShift(val)*12.0f;
		return sprintf(text, "%.1f notes", notes);
	}
} g_resample_fx;


//*************************************************************************************************
FFTFx *g_fft_fx_table[] =
{
	&g_filter_fx,
	&g_contrast_fx,
	&g_smear_fx,
	&g_thresh_fx,
	&g_clip_fx,
	&g_mirror_effect,
	&g_resample_fx,
	&g_shift_add_fx,
	&g_shift_addd_fx,
	&g_shift_repl_fx,
	&g_shift_repld_fx,
	&g_shift_move_fx,
	&g_shift_moved_fx,
	&g_harm_filt_fx,
	&g_auto_harm_fx, 
	&g_sweep1_fx,
	&g_sweep2_fx,
	&g_sweep3_fx,
	&g_sweep4_fx,
	&g_sweep5_fx,
	&g_harm_mask,
	&g_auto_harm_mask,
	&g_auto_subharm1_mask,
	&g_auto_subharm2_mask,
	&g_auto_subharm3_mask,
	&g_thresh_mask
};

const int g_num_fft_fx = NUM_ELEMENTS(g_fft_fx_table);

FFTFx* GetFFTFx(int idx)
{
	if(idx < 0 || idx >= g_num_fft_fx) return &g_no_fx;
	return g_fft_fx_table[idx];
}
