/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

Top-level DtBlkFx gui

Holds the following panels:
    GlobalCtrl
    2x spectrogram
    4x FxCtrl


History
	Date       Version    Programmer         Comments
	16/2/03    1.0        Darrell Tam		Created


This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/

#include <math.h>
#include <stdlib.h>	
#include <stdio.h>
#include <valarray>
#include <dt/windows_support.h>

#include "BlkFxParam.h"
#include "DTBlkFx.hpp"
#include "FFTFx.h"

#include "Spectrogram.h"
#include "Gui.h"

#include "FxCtrl.h"

#define DBG_FILE_NAME "c:\\gui.txt"
#include "Debug.h"

using namespace std;
using namespace VSTGUI;

enum
{
	// bitmaps
	// others
	kBackgroundW = 420,
	kBackgroundH = 410
};



//-------------------------------------------------------------------------------------------------
void Gui::generateAmpOverlay()
// generate exponential amplitude gfx into _amp_bm
// this code was written (instead of just using a normal bitmap) back when the whole gui was
// going to be scalable
{
	int width = _amp_bm->getWidth();
	int height = _amp_bm->getHeight();
	float fwidth = (float)width;
	float fheight = (float)height;

	// generate an exponential curve
	int zero_db = (int)(0.6f*fwidth);
	float xscale = 1.0f/zero_db;
	float yscale = 0.8f*fheight;
	CPoint p;
	for(p.x = 0; p.x < width; p.x++) {
		float top = yscale*(1-powf(20.0f, (float)p.x*xscale)/20.0f);

		int i_top = (int)floorf(top);
		for(p.y = 0; p.y < height; p.y++) {
			if(p.y < i_top) {
				_amp_bm->drawTransparent(p);
				continue;
			}
			unsigned alpha = 255;
			if(p.y == i_top) {
				// anti-alias
				float frac = (float)alpha*(1.0f-top+(float)i_top);
				alpha = (int)frac;
			}
			_amp_bm->drawPoint(p, 255, 0, 0, alpha);
		}
	}
}

//-------------------------------------------------------------------------------------------------
Gui::Gui (AudioEffect *effect) : AEffGUIEditor (effect)
{
	/*AEffGUIEditor*/frame = 0;
	oldTicks = 0;

	/*AEffGUIEditor::rect*/
	rect.left   = 0;
	rect.top    = 0;
	rect.right  = kBackgroundW;
	rect.bottom = kBackgroundH;

	// build the colour gradient map for the spectrograms
	float col_vec[][4] = {
	/*   pos     R     G     B */
		{0.0f,  0.0f, 0.0f, 0.0f},
		{0.2f,  0.0f, 0.0f, 1.0f},
		{0.4f,  0.0f, 1.0f, 1.0f},
		{0.6f,  0.0f, 0.8f, 0.0f},
		{0.8f,  1.0f, 1.0f, 0.0f},
		{1.0f,  1.0f, 0.0f, 0.0f}
	};
	_col_map.generate(1000, col_vec, 6);
	for(int i = 0; i < 2; i++) _sgram[i].set(_col_map);

	// we decide in this plugin to open all bitmaps in the open function
}

//-----------------------------------------------------------------------------
Gui::~Gui ()
{
	DBG("Gui::~Gui" << VAR(this));
}

//-----------------------------------------------------------------------------
long Gui::open (void *system_window)
// virtual, override AEffGUIEditor::open
// open the GUI window
//
{
	AEffGUIEditor::open (system_window);

	_pix_freq.build(
		400,						// width
		blkFx()->getSampleRate(),	// sample freq in Hz
		blkFx()->FFTdata().r		// fft data buffer is fixed
	);

	// get version
	int version = getVstGuiVersion ();
	int verMaj = (version & 0xFF00) >> 16;
	int verMin = (version & 0x00FF);

	CRect r;
	CPoint p;
	int i;

	// get bitmaps
	VstGuiRef<CBitmap> bg_bitmap;
	bg_bitmap.setNew(new CBitmap(10000));

	_freq_bg_bm.setNew(new CBitmap(10001));

	for(i = 0; i < 2; i++)
		_freq_knob_bm[i].setNew(new CBitmap(i+10100));

	_fxval_knob_bm.setNew(new CBitmap(10200));

	// create main window
	r(0, 0, 410, 410);
	frame = new CFrame (/*rectangle*/r, system_window, /*gui editor*/this);
	frame->setBackground (bg_bitmap);

	// place global controls
	p(0, 0);
	r(0, 0, _pix_freq.numPixels()+10, 15+13);
	if(_glob_ctrl.init(this, Offset(p, r))) frame->addView(&_glob_ctrl);

	// place spectrograms
	p.y = 28;
	r(4, 0, _pix_freq.numPixels()+2+4, 93);
	for(i = 0; i < 2; i++) {
		_sgram[i].init(
			Offset(p, r),		// total size
			_pix_freq			// pixel-freq map
		);
		frame->addView(_sgram+i);
		p.y += r.height()+1;
	}
	DBG("Gui::open, spectrograms ok");

	// how big each fx control panel is
	r(0, 0, _pix_freq.numPixels()+10, 45);

	// create amp overlay
	_amp_bm.setNew(new CAlphaContext(frame, r.width(), r.height()));
	if(_amp_bm && _amp_bm->ok()) generateAmpOverlay();
	DBG("Gui::open" << VAR(_amp_bm) << VAR(_amp_bm?_amp_bm->ok():false));

	// create fx ctrls
	p(0, 214);
	r(0, 0, _pix_freq.numPixels()+10, 48);
	for(i = 0; i < BlkFxParam::NUM_FX_SETS; i++) {
		if(_fx_ctrl[i].init(this, Offset(p, r))) frame->addView(_fx_ctrl+i);
		p.y += r.height();
	}

	// load params from the effect
	for(i = 0; i < BlkFxParam::TOTAL_NUM; i++) setParameter(i, blkFx()->getCurrParam(i));

	return true;
}

//-----------------------------------------------------------------------------
bool Gui::keysRequired ()
{
	if (frame && frame->getEditView()) return true;
	else return false;
}

//-----------------------------------------------------------------------------
long Gui::onKeyDown (VstKeyCode &keyCode)
{
	if (frame && 0 && (keyCode.character >= 'a' && keyCode.character <= 'z'))
	{
		char val[64];
		char modifiers[32];
		strcpy (modifiers, "");
		if (keyCode.modifier & MODIFIER_SHIFT)
			strcpy (modifiers, "Shift+");
		if (keyCode.modifier & MODIFIER_ALTERNATE)
			strcat (modifiers, "Alt+");
		if (keyCode.modifier & MODIFIER_COMMAND)
			strcat (modifiers, "Cmd+");
		if (keyCode.modifier & MODIFIER_CONTROL)
			strcat (modifiers, "Ctrl+");

		sprintf (val, "onKeyDown : '%s%c'", modifiers, (char)(keyCode.character));
		return 1;
	}

	if (frame && (keyCode.virt == VKEY_UP || keyCode.virt == VKEY_DOWN))
	{
		CView *pView = frame->getCurrentView ();
	}
	return -1;
}

//-----------------------------------------------------------------------------
long Gui::onKeyUp (VstKeyCode &keyCode)
{
	if (0 && (keyCode.character >= 'a' && keyCode.character <= 'z'))
	{
		char val[64];
		char modifiers[32];
		strcpy (modifiers, "");
		if (keyCode.modifier & MODIFIER_SHIFT)
			strcpy (modifiers, "Shift+");
		if (keyCode.modifier & MODIFIER_ALTERNATE)
			strcat (modifiers, "Alt+");
		if (keyCode.modifier & MODIFIER_COMMAND)
			strcat (modifiers, "Cmd+");
		if (keyCode.modifier & MODIFIER_CONTROL)
			strcat (modifiers, "Ctrl+");

		sprintf (val, "onKeyUp : '%s%c'", modifiers, (char)(keyCode.character));
		return 1;
	}

	return -1; 
}

//-----------------------------------------------------------------------------
void Gui::resume ()
{
	// called when the plugin will be On
}

//-----------------------------------------------------------------------------
void Gui::suspend ()
{
	// called when the plugin will be Off
	for(int i = 0; i < 2; i++) _sgram[i].suspend();
}

//-----------------------------------------------------------------------------
void Gui::close ()
// virtual, override AEffGUIEditor::close()
// the GUI window is closing
{
	delete /*AEffGUIEditor::*/frame;
	frame = 0;

}

//-----------------------------------------------------------------------------
void Gui::update(CDrawContext* context)
// virtual, override AEffGUIEditor
{
}

//-----------------------------------------------------------------------------
void Gui::idle ()
// virtual, override AEffGUIEditor::idle()
{
	//DBG("Gui::idle");
	AEffGUIEditor::idle ();		// always call this to ensure update
}

//-----------------------------------------------------------------------------
void Gui::setParameter (long index, float v)
// virtual, override AEffGUIEditor::setParameter()
//
// called from the AEffect to update the control's value
{
	using namespace BlkFxParam;
	BlkFxParam::SplitParamNum p(index);

	DBG("setParameter" << VAR(index) << VAR(v) << VAR(frame) << VAR(p.fx_set) << VAR(p.glob_param));

	// test if the plug is opened
	if (!frame) return;

	// pass parameter
	if(p.fx_set >= 0 && p.fx_set < 16)
		_fx_ctrl[p.fx_set].setParameter(index, v);

	else if(p.glob_param >= 0)
		_glob_ctrl.setParameter(index, v);

	// call this to be sure that the graphic will be updated
	postUpdate ();
}
