/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

Spectrogram window
The spectrogram (waterfall) displays FFT data. Colour is used to represent magnitude of power
in FFT. Each block is put at the bottom of the display and the display scrolls upwards.

This class expects to be given a PixelFreqMap by the owning class in order to determine which
FFT bins correspond to which pixels

Bugs:
Currently doesn't scroll correctly if other windows are on top of it
In the buzz version it scrolled properly by calling the windows ScrollWindow function

History
	Date       Version    Programmer         Comments
	16/2/03    -         Darrell Tam		Buzz version
	01/12/05    1.0        Darrell Tam		Ported to VST


This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/
#include <dt/misc_stuff.h>
#include <dt/NoteFreq.h>
#include "Spectrogram.h"

#define xxDBG_FILE_NAME "c:\\sgram.txt"
#include "Debug.h"
using namespace std;

//-------------------------------------------------------------------------------------------------
Spectrogram::Spectrogram() : CView(CRect(0,0,1,1))
{
	DBG("Spectrogram" << VAR(this));
	paused = false;

	tick_height = 1;
	frame_col = CColorRGB(255,255,255);

	// could try to get these from somewhere
	_pwr_min = 1e-4f * 1e-4f;
	_pwr_max = 0.2f * 0.2f;
	min_samps_per_line = 600;

	_ok = false;
}

Spectrogram::~Spectrogram()
{
	DBG("~Spectrogram" << VAR(this));
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::init(CRect size, PixelFreqMap& /*in*/ freq_map)
//
// note: expects the freq pixel mapping to be set first
//
{
	/*CView::*/setViewSize(size);

	_freq_map = &freq_map;
	_image_height = size.height();
	DBG("init" << VAR(_image_height) << VAR(freq_map.numPixels()));

	_vline_x[0] = _vline_x[1] = -1;
	_mouse.init();

	// clear out the image data
	_image_data.resize(_image_height*freq_map.numPixels());
	set_zero(_image_data);

	_curr_max.resize(freq_map.numPixels());
	_curr_line_is_empty = true;

	_line_samp_pos.resize(_image_height);
	set_zero(_line_samp_pos);

	_image_i = 0;
	_curr_line_samp_pos = 0;

	_disp_i[0] = _disp_i[1] = 0;

	// find client rectangle
	client_rect = size;
	client_rect.top++;
	client_rect.bottom -= /*Spectrogram::*/tick_height+1;
	client_rect.left = 0;
	client_rect.right = _freq_map->numPixels();
	client_rect = SetXCent(GetCent(size).x, client_rect);

	// mouse info is drawn here
	txt_rect = Offset(TopLeft(client_rect), CRect(5, 2, 120, 17));

	// when paused, msg is drawn here
	msg_rect = Offset(TopRight(client_rect), CRect(-120, 2, 0, 17));

	_focus = false;

	setDirty();

	if(_col_map) _ok = true;
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::suspend()
// call when sound suspends to reset sample position
{
	_curr_line_samp_pos = 0;
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::set(GradColMap& /*in*/ col_map)
{
	_col_map = &col_map;
	_col_log_offs = logf(_pwr_min);
	_col_log_scale = col_map.size()/(logf(_pwr_max)-_col_log_offs);
}

//-------------------------------------------------------------------------------------------------
inline unsigned long Spectrogram::mapCol(float in)
//
// given an input value use the 
//
{
	GradColMap& col_map = *_col_map;
	unsigned out;
	if(in > _pwr_min) {
		out = (unsigned)((logf(in)-_col_log_offs)*_col_log_scale+0.5f);
		if(out >= _col_map->size()) out = col_map.size()-1;
	}
	else out = 0;
	return col_map[out];
}

//-------------------------------------------------------------------------------------------------
inline void Spectrogram::getMaxVals(
	bool empty_line,	// whether this is a new line (pass as constant to help optimizer)
	int fft_idx,		// fft index
	float pwr_scale
)
//
// get maximum values for each pixel from bin ranges
{
	//DBG("getMaxVals" << VAR(empty_line) << VAR(fft_idx));
	// loop over the pixels to find max value in each bin range
	const pcplxf* br = _freq_map->pixToBinRng(fft_idx);

	float *curr_max = &_curr_max[0];
	float *curr_max_end = curr_max+_freq_map->numPixels();

	pcplxf tt = *br;

	// find max power in each pixel range
	pcplxf n = *br++;		// get first bin range
	while(curr_max != curr_max_end) {
		pcplxf r = n;		// current bin range
		n = *br++;			// next bin range

		// always do at least one bin for every output pixel
		float v = norm(*r++)*pwr_scale;
		float max_v;
		if(!empty_line) {
			max_v = *curr_max;
			if(v > max_v) max_v = v;
		}
		else {
			max_v = v;
		}

		// check any other bins up to the next bin position
		while(r.r < n.r) {
			float v = norm(*r++)*pwr_scale;
			if(v > max_v) max_v = v;
		}
		*curr_max++ = max_v;
	}
}

//------------------------------------------------------------------------------------------
void Spectrogram::newData(
	long samp_pos,		// sample position
	long time_fft_n,	// number of samples in time domain (may be less samples than fft freq data)
	int fft_idx,		// fft index as per freq_map
	float pwr_scale			// scaling to apply to data
)
// add some new data to the spectrogram
{
//	DBG("newData" << VAR(samp_pos) << VAR(time_fft_n) << VAR(fft_idx));
	if(paused || !_ok) return;

	if(!_curr_line_is_empty) {
		getMaxVals(/*empty_line*/false, fft_idx, pwr_scale);
	}
	else {
		_curr_line_is_empty = false;
		_line_samp_pos[_image_i] = samp_pos;
		getMaxVals(/*empty_line*/true, fft_idx, pwr_scale);
	}

	// don't display the line until we have processed enough samples
	if(samp_pos+time_fft_n-_curr_line_samp_pos < /*Spectrogram::*/min_samps_per_line) return;

	int w = _freq_map->numPixels();

	// render spectrogram line from the "curr_max" array into current line "_image_i"
	unsigned long *image_data = &_image_data[_image_i*w];

	float *curr_max = &_curr_max[0];
	float *curr_max_end = curr_max+w;
	while(curr_max != curr_max_end)
		*image_data++ = mapCol(*curr_max++);

	// current line sample position
	_curr_line_samp_pos = samp_pos+time_fft_n;

	// this line is ready to be displayed
	_disp_i[1] = _image_i;

	// circular buffer
	_image_i++;
	if(_image_i >= _image_height) _image_i = 0;

	_curr_line_is_empty = true;
}


//-------------------------------------------------------------------------------------------------
void Spectrogram::draw (
	CDrawContext *context,	//
	CRect clip_r			// limit drawing to this clipping rectangle
)
//
// draw spectrogram bitmap
//
// use _disp_i[0] as the index of most recent line in image buffer
{
	DBG("draw" << VAR(clip_r));
	context->setDrawMode(kCopyMode);

	clip_r.bound(client_rect);

	// make sure that we aren't trying to draw more than what we have
	if(client_rect.bottom - clip_r.top > _image_height) {
		CRect r = clip_r;

		// adjust the output rectangle
		clip_r.top = client_rect.bottom - _image_height;

		// fill rectangle at top for which there is no data with black
		r.bottom = clip_r.top;
		if(r.height() > 0) {
			context->setFillColor(kBlackCColor);
			context->fillRect(r);
		}
	}

	// create a memory dc containing the image data
	HBITMAP bm = CreateBitmap(
		_freq_map->numPixels(),		// width
		_image_height,				// height
		1,							// planes per pixel
		32,							// bits per pixel (actually 24 bits with 8 bits wasted)
		&_image_data[0]				// contiguous image data
	);

	CRect r = clip_r;

	// we need the windows drawing context
	HDC dc = (HDC)context->getSystemContext();
	HDC mdc = CreateCompatibleDC(dc);
	HGDIOBJ old_bm = SelectObject (mdc, bm);

	// perform copy in at most 2 blt's (depending on whether we need to wrap in src buffer)
	//
	int src_x = r.left-client_rect.left;

	while(1) {
		int n_lines = r.height();
		if(n_lines <= 0) break;

		// start line into image bitmap
		int src_i = 1+_disp_i[0] - (client_rect.bottom-r.top);
		bool temp = false;

		if(src_i < 0) {
			// wrap to end of source buffer
			src_i += _image_height;

			// limit number of lines to end-of-buffer
			int max_lines = _image_height-src_i;
			if(max_lines < n_lines) n_lines = max_lines;
			temp = true;
		}
		// do the copy
		DBG("draw" << VAR(r) << VAR(n_lines) << VAR(src_x) << VAR(src_i));
		BitBlt(
			dc,					// destination dc
			r.left,		// dest x
			r.top,			// dest y
			r.width(),		// width
			n_lines,			// height
			mdc,				// src dc
			src_x,				// src x
			src_i,				// src y
			SRCCOPY				// flags
		);
		r.top += n_lines;
	}
	DeleteObject(bm);
	DeleteDC(mdc);
}

//-------------------------------------------------------------------------------------------------
bool Spectrogram::writeMsg(CDrawContext *context)
{
	if(!paused) return false;

	const char* str = "disabled";

	context->setFont(kNormalFont);
	context->setFillColor(CColorRGB(255,255,255));
	context->setFontColor(CColorRGB(0,0,0));

	// draw some text
	OutlineDrawString(
		context,		// draw context
		msg_rect,		// rectangle
		str,			// string
		kCenterText		// horiz alignment
	);
	return true;
}

//-------------------------------------------------------------------------------------------------
bool Spectrogram::drawVLine(CDrawContext *context)
// draw a vertical line in the client context (or do nothing if vline_x < 0) using _vline_x[0]
{
	if(_vline_x[0] < 0) return false;

	CRect r = client_rect;
	r.left += _vline_x[0];
	r.right = r.left;
	r.bottom -= 1;
	FullInvertRect(context, r);
	return true;
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::update(CDrawContext *context)
// virtual, override CView::update
//
// called periodically, performs scrolling if need be and displays mouse information
{
	_mouse.update(context);

	if(_mouse.IsInside(size)) {
		// mouse is inside our rectangle
		setVLinePix(_mouse.pos.x-client_rect.left);
		_focus = true;
	}
	else if(_focus) {
		// mouse has gone out of focus
		clrVLine();
		_focus = false;
	}

	bool vline_moved = (_vline_x[1] != _vline_x[0]);

	// this is where we want to display to
	int new_disp_i = _disp_i[1];
	// note: _disp_i[1] may be updated asynchronously to this function - ensure code is multi-thread safe

	// check whether we need to scroll
	int scroll_n = new_disp_i-_disp_i[0];
	if(scroll_n < 0) scroll_n += _image_height;

	DBG("update" << VAR(this) <<  VAR(bDirty) << VAR(scroll_n) << VAR(_disp_i[0]) << VAR(new_disp_i) << VAR(vline_moved) << VAR(_vline_x[0]) << VAR(_vline_x[1]));

	// erase previous text & line
	if(scroll_n || vline_moved) {
		// erase text if there was a vline
		if(drawVLine(context)) draw(context, txt_rect);
	}

	// scroll upwards
	if(scroll_n) {
		CRect r = client_rect;
		r.bottom -= scroll_n;

		HDC dc = (HDC)context->getSystemContext();
		BitBlt(
			dc,					// destination dc
			r.left,				// dest x
			r.top,				// dest y
			r.width(),			// width
			r.height(),			// height
			dc,					// src dc
			r.left,				// src x
			r.top+scroll_n,		// src y
			SRCCOPY				// flags
		);

		// update the index to most recent line
		_disp_i[0] = new_disp_i;

		// draw in missing data
		r.top = r.bottom;
		r.bottom = client_rect.bottom;
		draw(context, r);
	}


	// draw the veritcal line & info text if need be
	_vline_x[0] = _vline_x[1];
	if((scroll_n || vline_moved) && drawVLine(context)) {

		context->setFont(kNormalFont);
		context->setFillColor(CColorRGB(255,255,255));
		context->setFontColor(CColorRGB(0,0,0));

		// draw some text
		char str[256], *s;
		float hz = _freq_map->pixToFreq(_vline_x[0]);
		s = str;
		s += HzToNote(s, hz);
		OutlineDrawString(
			context,		// draw context
			LeftHalf(/*Spectrogram::*/txt_rect),// rectangle
			str,			// string
			kCenterText		// horiz alignment
		);

		s = str;
		s += sprintNum(s, hz);
		s += sprintf(s, "Hz");
		OutlineDrawString(
			context,		// draw context
			RightHalf(/*Spectrogram::*/txt_rect),// rectangle
			str,			// string
			kCenterText		// horiz alignment
		);
	}

}

//-------------------------------------------------------------------------------------------------
void Spectrogram::draw (CDrawContext *context)
// virtual override CView::draw
//
{
	DBG("draw all" << VAR(this) << VAR(client_rect) << VAR(size));

	// draw borders and ticks
	context->setFillColor(/*Spectrogram::*/frame_col);

	// top border
	CRect r = /*CView::*/size;
	r.right--;
	r.bottom = client_rect.top-1;
	FullFillRect(context, r);
	// bottom border
	r.top = client_rect.bottom;
	r.bottom = r.top;
	FullFillRect(context, r);
	// left edge
	r.top = size.top;
	r.bottom = client_rect.bottom;
	r.left = size.left;
	r.right = client_rect.left-1;
	FullFillRect(context, r);
	// right edge
	r.left = client_rect.right;
	r.right = size.right-1;
	FullFillRect(context, r);

	r.top = client_rect.bottom+1;
	r.bottom = size.bottom-1;

	// octave ticks
	for(float x=0; x < size.width(); x += _freq_map->pixPerOctave()) {
		r.left = (int)(x+0.5f)+client_rect.left;
		r.right = r.left;
		FullFillRect(context, r);
	}

	// draw full spectrogram using most current
	_disp_i[0] = _disp_i[1];
	draw(context, client_rect);

	// write a msg if there is one
	writeMsg(context);

	// do update stuff
	_vline_x[0] = -1;
	update(context);

	// don't clear dirty so that update() will be called periodically
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::mouse(CDrawContext *context, CPoint &where)
// virtual, override CView method
//
// corresponds to mouse button press
//
{
	_mouse.click(context, where, size);

	// any mouse button toggles pause
	if(_mouse.click_mask) {
		paused = !paused;

		// erase the vline
		drawVLine(context);

		// write the message or erase area if not paused
		if(!writeMsg(context)) draw(context, msg_rect);

		// put back the vline
		drawVLine(context);
	}
}

//-------------------------------------------------------------------------------------------------
bool Spectrogram::removed(CView* parent)
// virtual, override CView::removed
{
	DBG("removed");
	CView::removed(parent);

	// could do more cleanup here
	_ok = false;
	setParent(NULL);
	return true;
}
