#ifndef _DT_VST_GUI_SUPPORT_H
#define _DT_VST_GUI_SUPPORT_H
/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com
VST is a trademark of Steinberg Media Technologies GmbH

VST GUI support

A collection of classes and templates to help you work better using steinberg's VST GUI classes

To much stuff to list... please have a look through the file.

History
	Date       Version    Programmer         Comments
	16/2/03    1.0        Darrell Tam		Created


This code is free software

***************************************************************************************************/

#include "misc_stuff.h"
#include <vstgui.h>
#include <vstcontrols.h>

#ifdef _DBG
#include <fstream>
inline std::ostream& operator << (std::ostream& o, CRect r)
	{ return o << "[" << r.left <<"," << r.top <<" "<< r.width() <<"x"<< r.height() <<"]"; }
inline std::ostream& operator << (std::ostream& o, CPoint p)
	{ return o << "[" << p.x <<"," << p.y <<"]"; }
inline std::ostream& operator << (std::ostream& o, CColor c) 
{ 
	char buf[32];
	sprintf(buf, "#%02d%02d%02d", c.red, c.green, c.blue);
	return o << buf;
}
#endif


struct CColorRGB : public CColor
// CColor doesn't have the constructor
{
	CColorRGB(unsigned char r, unsigned char g, unsigned char b)
		{ red = r; green = g; blue = b; unused = 0; }
	CColorRGB(const CColor& src) { CColor::operator =(src); }
};

// various functions to convert between CRect & CPoint
inline CRect MoveTo(CPoint top_left, CRect r) { return CRect(top_left.x, top_left.y, r.width()+top_left.x, r.height()+top_left.y); }
inline CRect MoveToOrigin(CRect r) { return MoveTo(CPoint(0,0), r); }
inline CRect Offset(CPoint p, CRect r) { return r.offset(p.x, p.y); }
inline CPoint TopLeft(CRect r) { return CPoint(r.left, r.top); }
inline CPoint TopRight(CRect r) { return CPoint(r.right, r.top); }
inline CPoint BottomRight(CRect r) { return CPoint(r.right, r.bottom); }
inline CPoint BottomLeft(CRect r) { return CPoint(r.left, r.bottom); }

// some operators for cpoints
inline CPoint operator - (CPoint a) { return CPoint(-a.x, -a.y); }
inline CPoint operator + (CPoint a, CPoint b) { return CPoint(a.x+b.x, a.y+b.y); }
inline CPoint operator - (CPoint a, CPoint b) { return operator +(a, -b); }
inline CPoint& operator += (CPoint& a, CPoint b) { return a.offset(b.x, b.y); }
inline CPoint& operator -= (CPoint& a, CPoint b) { return operator +=(a, -b); }

// return rectangle "r" centered horizontal at cx
inline CRect SetXCent(int cx, CRect r) { int w=r.width(); r.left = cx-w/2; r.right = r.left+w; return r; }

// return rectangle "r" centered vertical at cy
inline CRect SetYCent(int cy, CRect r) { int h=r.height(); r.top = cy-h/2; r.bottom = r.top+h; return r; }

// return rectangle "r" centred at point "p"
inline CRect SetCent(CPoint p, CRect r) { return SetYCent(p.y, SetXCent(p.x, r)); }

// return centre point of rectangle
inline CPoint GetCent(CRect r) { return CPoint((r.left+r.right)/2, (r.top+r.bottom)/2); }

inline CRect LeftHalf(CRect r) { r.right = GetCent(r).x; return r; }
inline CRect RightHalf(CRect r) { r.left = GetCent(r).x; return r; }
inline CRect TopHalf(CRect r) { r.bottom = GetCent(r).y; return r; }
inline CRect BottomHalf(CRect r) { r.top = GetCent(r).y; return r; }

inline CRect GetRect(CBitmap* bitmap) { return CRect(0, 0, bitmap->getWidth(), bitmap->getHeight()); }
inline CRect GetRect(CView* view) { CRect r; view->getViewSize(r); return r; }

inline void FullFillRect(CDrawContext* context, CRect r) { r.top--; r.left--; r.right++; r.bottom++; context->fillRect(r); }


//-------------------------------------------------------------------------------------------------
struct MouseUpdate
//
// class to store mouse state when using vst gui library
{
	// constants
	enum { kAllButtons = kRButton|kLButton|kMButton };

	MouseUpdate() { init(); }

	void init()
	{
		pos(0, 0);
		top_left(0, 0);
		buttons = 0;
		raw_buttons = 0;
		release_mask = 0;
		carry_release_mask = false;
		mouse_moved = false;
	}

	long update(CDrawContext* context)
	// update mouse position and chk for the release of any mouse buttons
	// (note, this method will only not look for button presses)
	// (call from update())
	{
		CPoint new_pos;
		context->getMouseLocation(new_pos);
		new_pos -= top_left;
		mouse_moved = pos != new_pos;
		pos = new_pos;

		raw_buttons = context->getMouseButtons();

		if(!carry_release_mask) release_mask = 0;

		// look for buttons that have been released
		for(long i = 1; i <= kRButton; i<<=1) {
			if((buttons&i) && !(raw_buttons&i)) {
				buttons &= ~i;		// clear the bit corresponding to this button
				release_mask |= i;	// 
			}
		}
		carry_release_mask = false;
		return release_mask;
	}
	
	// return true if:
	// a button is down and the initial click was in rectangle or
	// mouse is inside rectangle and no mouse button is down
	// rectangle is relative to point set by "setTopLeft"
	bool IsInside(CRect r) const
	{	
		// return true if any processed button is down 
		if(buttons & kAllButtons) return true;

		// return false if any raw button is down (meaning that it wasn't clicked
		// within our rectangle)
		if(raw_buttons & kAllButtons) return false;

		// no button is down, return whether the current mouse position is within rectangle
		return r.pointInside(pos);
	}

	void click(CDrawContext* context, CPoint new_pos, CRect r/*relative to new_pos*/)
	// call on a mouse click event
	// (call from mouse())
	{
		click_mask = 0;
		CPoint old_pos = pos;
		pos = new_pos-top_left;
		mouse_moved = pos != old_pos;

		if(!r.pointInside(new_pos)) {
			// any click outside of "r" causes all buttons to be released
			if(!carry_release_mask) release_mask = 0;
			release_mask |= buttons & kAllButtons;
			carry_release_mask = true;
			buttons = 0;
			return;
		}

		// click is inside rectangle, see what ones have been pressed
		long new_buttons = context->getMouseButtons();
		for(long i = 1; i <= kRButton; i<<=1) {
			if((new_buttons&i) && !(buttons&i)) {
				click_mask |= i;
				buttons |= i;
			}
		}
	}

	// return true if any mouse button is pressed
	bool isAnyButtonDown(bool raw=false) const
		{ return (raw?raw_buttons:buttons)&kAllButtons ? true: false; }

	// auto cast to CPoint
	operator CPoint& () { return pos; }
	operator CPoint () const { return pos; }

	// current processed buttons (& shift/alt)
	long buttons;

	// raw mouse buttons (& shift/alt)
	long raw_buttons;

	// latest mouse position (updated by "click()" & "update()")
	CPoint pos;			

	// offset to apply to mouse position on "click()" & "update()"
	CPoint top_left;

	// true after "click()" & "update()"
	bool mouse_moved;

	// which buttons if any were released after calling "update()" - use as mouse-up event
	long release_mask; 

	// which buttons have just been pressed after calling "click()" - use as a mouse-down event
	long click_mask;

	// release mask carried over from "click()" into "update()"
	bool carry_release_mask;


};

//-------------------------------------------------------------------------------------------------
inline bool IsFocus(CView* view, const MouseUpdate& mouse)
//
// return true if the mouse is over the control with no button down
// OR if the view parent has this view as the current edit view
//
{
	return GetRect(view).pointInside(mouse) && !mouse.isAnyButtonDown(/*raw*/true)
		|| view->getParent()->getEditView()==view;
}

//-------------------------------------------------------------------------------------------------
inline void FullInvertRect(CDrawContext* context, CRect r)
{
	// windows only...
	RECT r_ = {r.left, r.top, r.right+1, r.bottom+1 };
	InvertRect((HDC)context->getSystemContext(), &r_);
}

//-------------------------------------------------------------------------------------------------
inline void* Select(CDrawContext* context, CBitmap* bitmap)
// select a bitmap into the context
{
	// windows only at the moment
	return SelectObject((HDC)context->getSystemContext(), (HBITMAP)bitmap->getHandle());
}

//-------------------------------------------------------------------------------------------------
struct VstGuiObj
//
// object that can be used as base class for reference counting
// has the same ref-count semantics as other vst gui objects (CView, CBitmap, etc)
// e.g.
//
// class MyClass : virtual public VstGuiObj // important to inherit virtual
// {
//    ...
// }
//
// now use "VstGuiRef<MyClass>" to hold pointers to your class
// 
{
	VstGuiObj() { nbReference = 1; }
	virtual ~VstGuiObj() {}

	void forget () const {
		if (nbReference > 0) {
			if (!--nbReference) delete this;
		}
	}
	void remember () const { nbReference++; }
	long getNbReference () const { return nbReference; }
	mutable long nbReference;
};

//-------------------------------------------------------------------------------------------------
template <class T> struct VstGuiRef : public _PtrBase<T>
//
// wrapper for ptrs to vst gui objects
//
// remember() and forget() methods are automatically called
//
// Unfortunately for newly created objects you have to choose whether to assign directly
// or use "setNew", depending on what you want to do. The reason is that vst gui objects
// are created with a reference count of 1 and some gui object methods that
// take an object don't increment the reference count (e.g. CFrame::addView()) while others
// do (e.g. CFrame::setBackground()).
//
// For methods like CFrame::setBackground(), pass the object into the reference count
// using "setNew" - i.e.
//
//		VstGuiRef<CBitmap> my_bitmap; // don't assign new CBitmap(...) directly
//		my_bitmap.setNew(new CBitmap(...));
//		frame->setBackground(my_bitmap);
//
// Whereas for addView...
//		VstGuiRef<CHorizontalSlider> my_slider = new CHorizontalSlider(...);
//		frame->addView(my_slider);
//
// As with any simple reference counting scheme, avoid circular references because the
// memory won't get deleted.
//
{
	VstGuiRef(const VstGuiRef& src) { set(src); }
	VstGuiRef(T* ptr_ = NULL) { set(ptr_); }
	~VstGuiRef() { set(NULL); }

	// internal set
	void set(T* new_ptr)
	{
		T* old_ptr = ptr;
		if(new_ptr) new_ptr->remember();
		if(old_ptr) old_ptr->forget();			// object may get deleted at this point
		ptr = new_ptr;
	}

	// use this method to assign newly created objects
	void setNew(T* newly_created_object)
	{
		set(newly_created_object);
		newly_created_object->forget();
	}

	// auto cast to compatible reference
	template<class T2> operator VstGuiRef<T2>() const { return (VstGuiRef<T2>(ptr)); }

	VstGuiRef& operator = (const VstGuiRef& src) { set(src); return *this; }
};

//-------------------------------------------------------------------------------------------------
template <class T> class VstGuiNoRefCount : public T
//
// disable reference counting for a vst gui object that that is a member variable or stack variable
//
// must have a constructor that takes no arguments
//
// be careful the owning object won't go out of scope before any objects that hold a reference
{
public:
	virtual void forget () {}
	virtual void remember () {}
	VstGuiNoRefCount() { nbReference = 100; }
	~VstGuiNoRefCount() { nbReference = 0; }
};

//-------------------------------------------------------------------------------------------------
inline bool AlphaBlend(
	CDrawContext* dst_context,
	CRect dst_rect,					// destination rectangle
	CDrawContext* src_context,		// must have alpha channel & data must have been correctly pre-scaled
	CRect src_rect,					// source rectangle
	unsigned char overall_alpha		// 255 for max
)
// alpha blend function using VST gui's function pointer to alphablend
{
	// get AlphaBlend out of vstgui.cpp
	extern BOOL (WINAPI *pfnAlphaBlend)(
		HDC hdcDest,                 // handle to destination DC
		int nXOriginDest,            // x-coord of upper-left corner
		int nYOriginDest,            // y-coord of upper-left corner
		int nWidthDest,              // destination fwdith
		int nHeightDest,             // destination fheight
		HDC hdcSrc,                  // handle to source DC
		int nXOriginSrc,             // x-coord of upper-left corner
		int nYOriginSrc,             // y-coord of upper-left corner
		int nWidthSrc,               // source fwdith
		int nHeightSrc,              // source fheight
		BLENDFUNCTION blendFunction  // alpha-blending function
	);
	if(!pfnAlphaBlend) return false;

	BLENDFUNCTION bf;				// structure for alpha blending
	bf.BlendOp = AC_SRC_OVER;
	bf.BlendFlags = 0;
	bf.AlphaFormat = AC_SRC_ALPHA;   // use source alpha 
	bf.SourceConstantAlpha = overall_alpha;   // use constant alpha
	return (*pfnAlphaBlend)(
		(HDC)dst_context->getSystemContext(), // handle to destination DC
		dst_rect.left,				// x-coord of upper-left corner
		dst_rect.top,				// y-coord of upper-left corner
		dst_rect.width(),			// destination wdith
		dst_rect.height(),			// destination height
		(HDC)src_context->getSystemContext(),    // handle to source DC
		src_rect.left,				// x-coord of upper-left corner
		src_rect.top,				// y-coord of upper-left corner
		src_rect.width(),			// source fwdith
		src_rect.height(),			// source fheight
		bf							// alpha-blending function
	) ? true:false;
}

//-------------------------------------------------------------------------------------------------
inline void OutlineDrawString(
	CDrawContext* context,
	CRect r,					// rectangle to place txt
	const char *str, //int str_len,
	CHoriTxtAlign horiz_align = kCenterText
	//int vert_align = 0	//0=top, 1=middle, 2=bottom
)
//
// use font colour for outside and fill colour for inside, this is not an efficient version
// can someone write a better one??
//
{
	// not very efficient but simple
	context->drawString(str, Offset(CPoint(-1, 0), r), /*opaque*/0, horiz_align);
	context->drawString(str, Offset(CPoint( 1, 0), r), /*opaque*/0, horiz_align);
	context->drawString(str, Offset(CPoint( 0,-1), r), /*opaque*/0, horiz_align);
	context->drawString(str, Offset(CPoint( 0, 1), r), /*opaque*/0, horiz_align);
	CColor font_col = context->getFontColor();
	context->setFontColor(context->getFillColor());
	context->drawString(str, r, /*opaque*/0, horiz_align);
	context->setFontColor(font_col);
}

//-------------------------------------------------------------------------------------------------
class CAlphaContext : public CDrawContext, public virtual VstGuiObj
//
// Similar to COffscreenContext except that bitmap created is always RGB+alpha channel with memory
// directly accessible
//
// Windows only... can some one port to mac pls?
//
{
protected:
	//---------------------------------------------------------------------------------------------
	void init()
	{
		_hbm = NULL;
		_hbm_prev = NULL;
	}

	//---------------------------------------------------------------------------------------------
	void create(int width, int height)
	// create an RGB+alpha bitmap
	// expect pSystemContext to have been filled in
	{
		init();
		_width = width;
		_height_minus_1 = height-1;
		HDC dc = (HDC)pSystemContext;
		if(!dc) return;

		clipRect = CRect(0, 0, width, height);

		BITMAPINFO bmi;
		memset(&bmi, 0, sizeof(BITMAPINFO));
		bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
		bmi.bmiHeader.biWidth = width;
		bmi.bmiHeader.biHeight = height;
		bmi.bmiHeader.biPlanes = 1;
		bmi.bmiHeader.biBitCount = 32;         // four 8-bit components
		bmi.bmiHeader.biCompression = BI_RGB;
		bmi.bmiHeader.biSizeImage = width * height * 4;

		void *data_ = NULL;
		_hbm = CreateDIBSection(dc, &bmi, DIB_RGB_COLORS, &data_, NULL, 0x0);
		if(!_hbm) return;

		_bitmap_data = (unsigned long*)data_;
		_hbm_prev = SelectObject(dc, _hbm);

		// flush any pending drawing operations
		GdiFlush();
	}

public:
	//---------------------------------------------------------------------------------------------
	CAlphaContext (
		CFrame* frame,	// frame to attach to
		int width,		// pixels
		int height		// pixels
	)
	: CDrawContext (
			frame,		// frame
			NULL,		// system context
			NULL		// window
		)
	{
		existNum()++;

		HWND wnd = (HWND)pFrame->getSystemWindow ();
		if(!wnd) return;

		HDC wnd_dc = GetDC(wnd);
		if(!wnd_dc) return;

		pSystemContext = CreateCompatibleDC (wnd_dc);

		ReleaseDC (wnd, wnd_dc);

		create(width, height);
	}


	//---------------------------------------------------------------------------------------------
	virtual ~CAlphaContext()
	{
		existNum()--;

		if(_hbm_prev) SelectObject((HDC)pSystemContext, _hbm_prev);
		DeleteObject(_hbm);

		if(pSystemContext) 
			DeleteDC((HDC)pSystemContext);

		pSystemContext = NULL;
	}

	//---------------------------------------------------------------------------------------------
	// return true if the bitmap memory was allocated ok
	virtual bool ok() { return _bitmap_data?true:false; }

	// return pointer to data with p(0,0) corresponding to top left hand corner
	unsigned long* data(const CPoint &p)
		{ return _bitmap_data + p.x + (_height_minus_1-p.y)*_width; }

	// raw draw point/get point
	void _drawPoint (const CPoint &p, unsigned long dat) { *(data(p)) = dat; }
	unsigned long _getPoint (const CPoint& p) { return *(data(p)); }
	void _drawPoint(const CPoint &p, unsigned char r, unsigned char g, unsigned char b, unsigned char a)
	{
		union {
			unsigned char d[4];
			unsigned long l;
		};
		d[3] = a; d[2] = r; d[1] = g; d[0] = b;
		_drawPoint(p, l);
	}

	// note: msdn says that you should call GdiFlush() before accessing bitmap memory directly
	// (all of these drawPoint methods access memory directly) after using any of the system drawing routines

	// draw completely opaque point (r,g,b values should be 0..255)
	void drawPoint(const CPoint &p, unsigned r, unsigned g, unsigned b)
		{ _drawPoint(p, (unsigned char)r, (unsigned char)g, (unsigned char)b, 255); }

	void drawPoint(const CPoint &p, const CColor& c)
		{ drawPoint(p, c.red, c.green, c.blue); }

	// draw point with transparency (r,g,b,a values should be 0..255)
	void drawPoint(const CPoint &p, unsigned r, unsigned g, unsigned b, unsigned a)
		{ _drawPoint(p, (unsigned char)((r * a)/255), (unsigned char)((g * a)/255), (unsigned char)((b * a)/255), (unsigned char)a); }

	void drawPoint(const CPoint &p, const CColor& c, unsigned a)
		{ drawPoint(p, c.red, c.green, c.blue, a); }

	// draw completely transparent point
	void drawTransparent(const CPoint &p)
		{ _drawPoint(p, 0); }


	//---------------------------------------------------------------------------------------------
	bool blend(
		CDrawContext* dst_context,
		CRect dst_rect,					// destination rectangle
		CRect src_rect,					// source rectangle
		unsigned char overall_alpha		// 255 for max
	)
	{ return AlphaBlend(dst_context, dst_rect, this, src_rect, overall_alpha); }

	//---------------------------------------------------------------------------------------------
	// non-scaling version, always copy from origin
	bool blend(
		CDrawContext* dst_context,
		CRect dst_rect,					// destination rectangle
		unsigned char overall_alpha		// 255 for max
	)
	{ return blend(dst_context, dst_rect, MoveToOrigin(dst_rect), overall_alpha); }

	inline long getWidth ()  { return _width; }

	inline long getHeight () { return _height_minus_1+1; }

	CRect getSize() { return CRect(0, 0, _width, _height_minus_1+1); }

	// for debugging, find number of these objects in existence
	static int& existNum() { static int x = 0; return x; }
protected:
	// bitmap width & height
	int _width, _height_minus_1;

	HBITMAP _hbm;
	HGDIOBJ _hbm_prev;

	Ptr<unsigned long> _bitmap_data;
};
inline CRect GetRect(CAlphaContext* t) { return CRect(0, 0, t->getWidth(), t->getHeight()); }



#endif
