/******************************************************************************
Miscellaneous stuff

History
	Date       Version    Programmer         Comments
	16/2/03    1.0        Darrell Tam		Created

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

******************************************************************************/

#ifndef _MISC_STUFF_H_
#define _MISC_STUFF_H_

#include <stdio.h>
#include <memory>
#include <vector>
#include <valarray>

// for debugging
#define VAR(x) ", " #x "=" << (x)

// stream "X" into std::string "S"
#define STRSTR(S, X) { std::ostringstream my_ostringstream; my_ostringstream << X; S = my_ostringstream.str(); }

// sizeof an array in elements
#define NUM_ELEMENTS(a) (sizeof(a)/sizeof(a[0]))

//------------------------------------------------------------------------------------------
//
// return linear interpolation at point <frac> between <out_min> &
// <out_max>, where <frac> is in the range 0 to 1
//
inline float lin_interp(float frac, float out_min, float out_max)
	{ return (1.0f-frac)*out_min + frac*out_max; }

//------------------------------------------------------------------------------------------
template<class A, class B, class C>
inline A limit_range(const A& in, const B& range_min, const C &range_max)
// limit range of "in" between "range_min" & "range_max"
{
	if(in < range_min) return range_min;
	if(in > range_max) return range_max;
	return in;
}

//------------------------------------------------------------------------------------------
template <class I, class X> inline I wrap(I i, const X& x)
// wrap index "i" inside of "x" that contains the method "size()" returning the number of
// elements
{
	I n = x.size();
	i %= n;
	if(i < 0) i += n;
	return i;
}

//------------------------------------------------------------------------------------------
inline long prbs29(long x)
// pseudo random binary sequence (this one uses 29 bits and is 234880995 long - I chose it
// at random - there are probably longer ones for the same complexity)
{
	return ((x<<1) ^ ((x>>22)&1) ^ ((x>>25)&1) ^ ((x>>28)&1));
}

//------------------------------------------------------------------------------------------
template <class T, int _SIZE> struct Array
// fixed length C array wrapper,
// advantage over normal declaration is that is assignment is allowed
{
	typedef T TYPE;
	// number of elements
	enum { SIZE = _SIZE };
	static int size() { return SIZE; }

	// data store
	T data[SIZE];

	// force deep-copy from a pointer
	void copy(const T* p) { *this = *(const Array*)p; }

	// array operator
	const T& operator[] (int i) const { return data[i]; }
	T& operator[] (int i) { return data[i]; }

	// cast to pointer
	operator T* () { return data; }
};

//------------------------------------------------------------------------------------------
template <class T, class BASE=T> struct new_auto_ptr : public std::auto_ptr<BASE>
// wrapper for auto_ptr that is initialized with a new instance of "T"
// i.e. instead of writing this:
//     auto_ptr<MyClass> my_class = new MyClass;
// use:
//     new_auto_ptr<MyClass> my_class;

{
	new_auto_ptr() : auto_ptr<BASE>(new T) { }
};

//------------------------------------------------------------------------------------------
template<class T> struct _PtrBase
//
{
	T* ptr;

	_PtrBase() { ptr = NULL; }

	// auto cast to ptr
	operator T* () const { return ptr; }

	// dereference operator
	T& operator * () const { return *ptr; }

	// array operator
	T& operator [] (int i) const { return ptr[i]; }

	// pointing at operator
	T* operator -> () const { return ptr; }
};

//------------------------------------------------------------------------------------------
template<class T> struct Ptr : public _PtrBase<T>
//
// pointer wrapper template
// pointer automatically set to NULL on creation
// Delete/DeleteArray methods clear the pointer
//
// Works like a normal C++ pointer in that there is no concept of object ownership
//
{
	Ptr(T* ptr_ = NULL) { ptr = ptr_; }

	// auto cast to compatible ptr
	template<class T2> operator Ptr<T2>() const { return (Ptr<T2>(ptr)); }

	// delete and clear pointer
	void Delete() { delete ptr; ptr = NULL; }

	void DeleteArray() { delete [] ptr; ptr = NULL; }

	void New() { ptr = new T; }

	void NewArray(unsigned i) { ptr = new T[i]; }

};

//------------------------------------------------------------------------------------------
template <class T> inline void set_zero(std::vector<T> &v) { memset(&v[0], 0, sizeof(T)*v.capacity()); }
template <class T> inline void set_zero(std::valarray<T> &v) { memset(&v[0], 0, sizeof(T)*v.size()); }

// allow an auto_ptr to be set to a ptr
template<class A, class B> set(std::auto_ptr<A>& a, B* b) { a = std::auto_ptr<A>(b); }



//------------------------------------------------------------------------------------------
struct sprintNum
{
	// num chars printed
	int n;
	float min_unit;
	float max_unit;
	float abs_v;

	bool less_than(float unit)
		{ return (abs_v < unit && unit > min_unit) || (unit >= max_unit); }

	sprintNum(char* s, float v, float min_unit_=0.0f, float max_unit_=1e10f)
	// print a value with 3 significant figures and use append u (micro), m (milli) or k (kilo)
	{
		abs_v = fabsf(v);
		min_unit = min_unit_;
		max_unit = max_unit_;
		     if(less_than(0.00001f)) n = sprintf(s, "%.2fu", v*1e6f);		// micro
		else if(less_than(0.0001f)) n = sprintf(s, "%.1fu", v*1e6f);			// micro
		else if(less_than(0.001f)) n = sprintf(s, "%.0fu", v*1e6f);			// micro
		else if(less_than(0.01f)) n = sprintf(s, "%.2fm", v*1e3f);			// milli
		else if(less_than(0.1f)) n = sprintf(s, "%.1fm", v*1e3f);			// milli
		else if(less_than(1.0f)) n = sprintf(s, "%.0fm", v*1e3f);			// milli
		else if(less_than(10.0f)) n = sprintf(s, "%.2f", v);					// no multiplier
		else if(less_than(100.0f)) n = sprintf(s, "%.1f", v);				// no multiplier
		else if(less_than(1000.0f)) n = sprintf(s, "%.0f", v);				// no multiplier
		else if(less_than(10000.0f)) n = sprintf(s, "%.2fk", v*1e-3f);		// kilo
		else if(less_than(100000.0f)) n = sprintf(s, "%.1fk", v*1e-3f);	// kilo
		else                      n = sprintf(s, "%.0fk", v*1e-3f);
	}

	// auto cast to int
	operator int () { return n; }
};

//------------------------------------------------------------------------------------------
struct FixPoint
//
// fixed point arithmetic is 64 bit precision: 32 bits integer (signed) and 31 bits fraction
// use in place of doubles or floats if you just adding/subtracting and actually want
// long integers
//
{
	// whole part
	long i;

	// fractional part (0=+0, 0x7fffffff=+0.99999...)
	long frac;

	FixPoint() {}

	FixPoint(long i_, long frac_=0)
	{
		i = i_;
		frac = frac_;
	}

	FixPoint(int i_) 
	{
		i = i_;
		frac = 0;
	}

	FixPoint(double f)
	{
		double fi = floor(f);
		i = (long)fi;
		frac = (long)((f-fi)*0x80000000+0.5f);
	}

	// auto cast & round to a long
	operator long () const
	{
		if(frac >= 0x40000000 /*0.5*/) return i+1;
		return i;
	}

	// auto cast to double
	operator double () const
	{
		return (double)i+(double)frac/0x80000000;
	}
};

// negative operator
inline FixPoint operator - (FixPoint f)
	{ return f.frac?FixPoint(-f.i-1, 0x80000000-f.frac) : FixPoint(-f.i); }

// addition operator
inline FixPoint operator + (FixPoint f1, FixPoint f2)
{
	long i = f1.i + f2.i;
	long frac = f1.frac + f2.frac;
	if(frac < 0) {
		// see if the fraction overflowed
		frac -= 0x80000000;
		i ++;
	}
	return FixPoint(i, frac);
}
// derived operators
inline FixPoint operator - (FixPoint f1, FixPoint f2) { return operator +(f1, -f2); }
inline FixPoint& operator += (FixPoint& f1, FixPoint f2) { return f1 = f1+f2; }
inline FixPoint& operator -= (FixPoint& f1, FixPoint f2) { return f1 = f1-f2; }

// stream operator
#ifdef _OSTREAM_
inline std::ostream& operator << (std::ostream& o, FixPoint fix)
{
	char buf[64];
	sprintf(buf, " (%d:%08x)", fix.i, fix.frac);
	return o << (double)fix << buf;
}
#endif

//------------------------------------------------------------------------------------------
template <class T> struct FindMinMax
{
	T min_v, max_v;
	FindMinMax(T min_v_, T max_v_)
	{
		min_v = min_v_;
		max_v = max_v_;
	}
	void operator () (T v)
	{
		if(v < min_v) min_v = v;
		if(v > max_v) max_v = v;
	}
};
#ifdef _OSTREAM_
template<class T> inline std::ostream& operator << (std::ostream& o, FindMinMax<T> mm)
{
	return o << mm.min_v << ":" << mm.max_v;
}
#endif


//*************************************************************************************************
template <int _N> struct SplitParam
//
// split a param (range 0..1) into an integer part (range 0..N-1) and a fraction (range 0..1)
//
{
	enum { N = _N };

	int i_part;	// 0 .. N-1
	float f_part; // 0 .. 0.9999f

	SplitParam() {}
	SplitParam(float/*0..1*/param) { operator () (param); }

	void operator () (float/*0..1*/param)
	// divide the effect param into 2 parts
	{
		// special case
		if(param >= 1.0f) {
			i_part = N-1;
			f_part = 1.0f;
			return;
		}
		// normal case
		float i;
		param *= (float)N;
		f_part = modff(param, &i);
		i_part = limit_range((int)i, 0, N-1);
	}

	operator float ()
	// combine 2 params
	{
		float r = limit_range((float)i_part, 0.0f, (float)(N-1));
		r += limit_range(f_part, 0.0f, 0.9999f);
		return r* (1.0f/(float)N);
	}

};

#endif
